# RUN: %PYTHON %s 2>&1 | FileCheck %s

from mlir.ir import *
from mlir.dialects import arith
from mlir.dialects.ext import *
from mlir import ir
from typing import Any, Optional, Sequence, TypeVar, Union
import sys


def run(f):
    print("\nTEST:", f.__name__)
    f()


# CHECK: TEST: testMyInt
@run
def testMyInt():
    class MyInt(Dialect, name="myint"):
        pass

    i32 = IntegerType[32]

    class ConstantOp(MyInt.Operation, name="constant"):
        value: IntegerAttr
        cst: Result[i32] = result(infer_type=True)

    class AddOp(Operation, dialect=MyInt, name="add"):
        lhs: Operand[i32]
        rhs: Operand[i32]
        res: Result[i32] = result(infer_type=True)

    # CHECK: irdl.dialect @myint {
    # CHECK:   irdl.operation @constant {
    # CHECK:     %0 = irdl.base "#builtin.integer"
    # CHECK:     irdl.attributes {"value" = %0}
    # CHECK:     %1 = irdl.is i32
    # CHECK:     irdl.results(cst: %1)
    # CHECK:   }
    # CHECK:   irdl.operation @add {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(lhs: %0, rhs: %0)
    # CHECK:     irdl.results(res: %0)
    # CHECK:   }
    # CHECK: }
    with Context():
        MyInt.load()
        print(MyInt._mlir_module)

        # CHECK: ['constant', 'add']
        print([i._op_name for i in MyInt.operations])
        i32 = IntegerType.get_signless(32)

        with Location.unknown():
            module = Module.create()
            with InsertionPoint(module.body):
                two = ConstantOp(IntegerAttr.get(i32, 2))
                three = ConstantOp(IntegerAttr.get(i32, 3))
                add1 = AddOp(two, three)
                add2 = AddOp(add1, two)
                add3 = AddOp(add2, three)

        # CHECK: %0 = "myint.constant"() {value = 2 : i32} : () -> i32
        # CHECK: %1 = "myint.constant"() {value = 3 : i32} : () -> i32
        # CHECK: %2 = "myint.add"(%0, %1) : (i32, i32) -> i32
        # CHECK: %3 = "myint.add"(%2, %0) : (i32, i32) -> i32
        # CHECK: %4 = "myint.add"(%3, %1) : (i32, i32) -> i32
        print(module)
        assert module.operation.verify()

        # CHECK: AddOp
        print(type(add1).__name__)
        # CHECK: ConstantOp
        print(type(two).__name__)
        # CHECK: myint.add
        print(add1.OPERATION_NAME)
        # CHECK: None
        print(add1._ODS_OPERAND_SEGMENTS)
        # CHECK: None
        print(add1._ODS_RESULT_SEGMENTS)
        # CHECK: (0, True)
        print(add1._ODS_REGIONS)
        # CHECK: %0 = "myint.constant"() {value = 2 : i32} : () -> i32
        print(add1.lhs.owner)
        # CHECK: %1 = "myint.constant"() {value = 3 : i32} : () -> i32
        print(add1.rhs.owner)
        # CHECK: 2 : i32
        print(two.value)
        # CHECK: OpResult(%0
        print(two.cst)
        # CHECK: (self, /, lhs, rhs, *, res=None, loc=None, ip=None)
        print(AddOp.__init__.__signature__)
        # CHECK: (self, /, value, *, cst=None, loc=None, ip=None)
        print(ConstantOp.__init__.__signature__)

        # CHECK: True
        print(issubclass(AddOp.Adaptor, OpAdaptor))
        adaptor1 = AddOp.Adaptor(list(add1.operands), add1)
        # CHECK: myint.add
        print(adaptor1.OPERATION_NAME)
        # CHECK: OpResult(%0 = "myint.constant"() {value = 2 : i32} : () -> i32)
        print(adaptor1.lhs)
        # CHECK: OpResult(%1 = "myint.constant"() {value = 3 : i32} : () -> i32)
        print(adaptor1.rhs)


# CHECK: TEST: testExtDialect
@run
def testExtDialect():
    class Test(Dialect, name="ext_test"):
        pass

    i32 = IntegerType[32]

    class ConstraintOp(Test.Operation, name="constraint"):
        a: Operand[i32 | IntegerType[64]]
        b: Operand[Any]
        # Here we use `F32Type[()]` instead of just `F32Type`
        # because of an existing issue in IRDL implementation
        # where `irdl.base` cannot exist in `irdl.any_of`.
        c: Operand[F32Type[()] | i32]
        d: Operand[Any]
        x: IntegerAttr
        y: FloatAttr

    class OptionalOp(Test.Operation, name="optional"):
        out1: Result[i32]
        out3: Result[i32]
        a: Operand[i32]
        out2: Result[i32] | None = None
        b: Optional[Operand[i32]] = None

    class Optional2Op(Test.Operation, name="optional2"):
        b: Optional[Result[i32]] = None
        a: Optional[Operand[i32]] = None

    class VariadicOp(Test.Operation, name="variadic"):
        out1: Sequence[Result[i32]]
        out2: Sequence[Result[i32]]
        out4: Result[i32]
        a: Operand[i32]
        c: Sequence[Operand[i32]]
        out3: Optional[Result[i32]] = None
        b: Optional[Operand[i32]] = None

    class Variadic2Op(Test.Operation, name="variadic2"):
        b: Sequence[Result[i32]]
        a: Sequence[Operand[i32]]

    class MixedOpBase(Test.Operation):
        out: Result[i32]
        in1: Operand[i32]

    class MixedOp(MixedOpBase, name="mixed"):
        in2: IntegerAttr
        in4: IntegerAttr
        in5: Operand[i32]
        in3: Optional[Operand[i32]] = None

    T = TypeVar("T")
    U = TypeVar("U", bound=IntegerType[32] | IntegerType[64])
    V = TypeVar("V", bound=Union[IntegerType[8], IntegerType[16]])

    class TypeVarOp(Test.Operation, name="type_var"):
        in1: Operand[T]
        in2: Operand[T]
        in3: Operand[U]
        in4: Operand[U | V]
        in5: Operand[V]

    class OptionalButNotKeywordOp(Test.Operation, name="optional_but_not_keyword"):
        a: Operand[i32]
        b: Optional[Operand[i32]]
        c: Operand[i32]

    # CHECK: irdl.dialect @ext_test {
    # CHECK:   irdl.operation @constraint {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     %1 = irdl.is i64
    # CHECK:     %2 = irdl.any_of(%0, %1)
    # CHECK:     %3 = irdl.any
    # CHECK:     %4 = irdl.is f32
    # CHECK:     %5 = irdl.any_of(%4, %0)
    # CHECK:     %6 = irdl.any
    # CHECK:     irdl.operands(a: %2, b: %3, c: %5, d: %6)
    # CHECK:     %7 = irdl.base "#builtin.integer"
    # CHECK:     %8 = irdl.base "#builtin.float"
    # CHECK:     irdl.attributes {"x" = %7, "y" = %8}
    # CHECK:   }
    # CHECK:   irdl.operation @optional {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(a: %0, b: optional %0)
    # CHECK:     irdl.results(out1: %0, out3: %0, out2: optional %0)
    # CHECK:   }
    # CHECK:   irdl.operation @optional2 {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(a: optional %0)
    # CHECK:     irdl.results(b: optional %0)
    # CHECK:   }
    # CHECK:   irdl.operation @variadic {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(a: %0, c: variadic %0, b: optional %0)
    # CHECK:     irdl.results(out1: variadic %0, out2: variadic %0, out4: %0, out3: optional %0)
    # CHECK:   }
    # CHECK:   irdl.operation @variadic2 {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(a: variadic %0)
    # CHECK:     irdl.results(b: variadic %0)
    # CHECK:   }
    # CHECK:   irdl.operation @mixed {
    # CHECK:     %0 = irdl.is i32
    # CHECK:     irdl.operands(in1: %0, in5: %0, in3: optional %0)
    # CHECK:     %1 = irdl.base "#builtin.integer"
    # CHECK:     %2 = irdl.base "#builtin.integer"
    # CHECK:     irdl.attributes {"in2" = %1, "in4" = %2}
    # CHECK:     irdl.results(out: %0)
    # CHECK:   }
    # CHECK:   irdl.operation @type_var {
    # CHECK:     %0 = irdl.any
    # CHECK:     %1 = irdl.is i32
    # CHECK:     %2 = irdl.is i64
    # CHECK:     %3 = irdl.any_of(%1, %2)
    # CHECK:     %4 = irdl.is i8
    # CHECK:     %5 = irdl.is i16
    # CHECK:     %6 = irdl.any_of(%4, %5)
    # CHECK:     %7 = irdl.any_of(%3, %6)
    # CHECK:     irdl.operands(in1: %0, in2: %0, in3: %3, in4: %7, in5: %6)
    # CHECK:   }
    # CHECK: }
    with Context(), Location.unknown():
        Test.load()
        print(Test._mlir_module)

        # CHECK: (self, /, a, b, c, d, x, y, *, loc=None, ip=None)
        print(ConstraintOp.__init__.__signature__)
        # CHECK: (self, /, out1, out3, a, *, out2=None, b=None, loc=None, ip=None)
        print(OptionalOp.__init__.__signature__)
        # CHECK: (self, /, *, b=None, a=None, loc=None, ip=None)
        print(Optional2Op.__init__.__signature__)
        # CHECK: (self, /, out1, out2, out4, a, c, *, out3=None, b=None, loc=None, ip=None)
        print(VariadicOp.__init__.__signature__)
        # CHECK: (self, /, b, a, *, loc=None, ip=None)
        print(Variadic2Op.__init__.__signature__)
        # CHECK: (self, /, out, in1, in2, in4, in5, *, in3=None, loc=None, ip=None)
        print(MixedOp.__init__.__signature__)
        # CHECK: (self, /, in1, in2, in3, in4, in5, *, loc=None, ip=None)
        print(TypeVarOp.__init__.__signature__)
        # CHECK: (self, /, a, b, c, *, loc=None, ip=None)
        print(OptionalButNotKeywordOp.__init__.__signature__)

        # CHECK: None None
        print(ConstraintOp._ODS_OPERAND_SEGMENTS, ConstraintOp._ODS_RESULT_SEGMENTS)
        # CHECK: [1, 0] [1, 1, 0]
        print(OptionalOp._ODS_OPERAND_SEGMENTS, OptionalOp._ODS_RESULT_SEGMENTS)
        # CHECK: [0] [0]
        print(Optional2Op._ODS_OPERAND_SEGMENTS, Optional2Op._ODS_RESULT_SEGMENTS)
        # CHECK: [1, -1, 0] [-1, -1, 1, 0]
        print(VariadicOp._ODS_OPERAND_SEGMENTS, VariadicOp._ODS_RESULT_SEGMENTS)
        # CHECK: [-1] [-1]
        print(Variadic2Op._ODS_OPERAND_SEGMENTS, Variadic2Op._ODS_RESULT_SEGMENTS)

        i32 = IntegerType.get_signless(32)
        i64 = IntegerType.get_signless(64)
        f32 = F32Type.get()

        iattr = IntegerAttr.get(i32, 2)
        fattr = FloatAttr.get_f32(2.3)

        module = Module.create()
        with InsertionPoint(module.body):
            ione = arith.constant(i32, 1)
            fone = arith.constant(f32, 1.2)

            # CHECK: "ext_test.constraint"(%c1_i32, %c1_i32, %cst, %c1_i32) {x = 2 : i32, y = 2.300000e+00 : f32} : (i32, i32, f32, i32) -> ()
            c1 = ConstraintOp(ione, ione, fone, ione, iattr, fattr)
            # CHECK: "ext_test.constraint"(%c1_i32, %cst, %cst, %cst) {x = 2 : i32, y = 2.300000e+00 : f32} : (i32, f32, f32, f32) -> ()
            ConstraintOp(ione, fone, fone, fone, iattr, fattr)
            # CHECK: ext_test.constraint"(%c1_i32, %cst, %c1_i32, %cst) {x = 2 : i32, y = 2.300000e+00 : f32} : (i32, f32, i32, f32) -> ()
            ConstraintOp(ione, fone, ione, fone, iattr, fattr)

            # CHECK: %0:2 = "ext_test.optional"(%c1_i32) {operandSegmentSizes = array<i32: 1, 0>, resultSegmentSizes = array<i32: 1, 1, 0>} : (i32) -> (i32, i32)
            o1 = OptionalOp(i32, i32, ione)
            # CHECK: %1:3 = "ext_test.optional"(%c1_i32, %c1_i32) {operandSegmentSizes = array<i32: 1, 1>, resultSegmentSizes = array<i32: 1, 1, 1>} : (i32, i32) -> (i32, i32, i32)
            o2 = OptionalOp(i32, i32, ione, out2=i32, b=ione)
            # CHECK: ext_test.optional2"() {operandSegmentSizes = array<i32: 0>, resultSegmentSizes = array<i32: 0>} : () -> ()
            o3 = Optional2Op()
            # CHECK: %2 = "ext_test.optional2"() {operandSegmentSizes = array<i32: 0>, resultSegmentSizes = array<i32: 1>} : () -> i32
            o4 = Optional2Op(b=i32)
            # CHECK: "ext_test.optional2"(%c1_i32) {operandSegmentSizes = array<i32: 1>, resultSegmentSizes = array<i32: 0>} : (i32) -> ()
            o5 = Optional2Op(a=ione)
            # CHECK: %3 = "ext_test.optional2"(%c1_i32) {operandSegmentSizes = array<i32: 1>, resultSegmentSizes = array<i32: 1>} : (i32) -> i32
            o6 = Optional2Op(b=i32, a=ione)

            # CHECK: %4:4 = "ext_test.variadic"(%c1_i32, %c1_i32, %c1_i32) {operandSegmentSizes = array<i32: 1, 2, 0>, resultSegmentSizes = array<i32: 1, 2, 1, 0>} : (i32, i32, i32) -> (i32, i32, i32, i32)
            v1 = VariadicOp([i32], [i32, i32], i32, ione, [ione, ione])
            # CHECK: %5:5 = "ext_test.variadic"(%c1_i32, %c1_i32, %c1_i32) {operandSegmentSizes = array<i32: 1, 1, 1>, resultSegmentSizes = array<i32: 1, 2, 1, 1>} : (i32, i32, i32) -> (i32, i32, i32, i32, i32)
            v2 = VariadicOp([i32], [i32, i32], i32, ione, [ione], out3=i32, b=ione)
            # CHECK: %6:4 = "ext_test.variadic"(%c1_i32) {operandSegmentSizes = array<i32: 1, 0, 0>, resultSegmentSizes = array<i32: 2, 1, 1, 0>} : (i32) -> (i32, i32, i32, i32)
            v3 = VariadicOp([i32, i32], [i32], i32, ione, [])
            # CHECK: "ext_test.variadic2"() {operandSegmentSizes = array<i32: 0>, resultSegmentSizes = array<i32: 0>} : () -> ()
            v4 = Variadic2Op([], [])
            # CHECK: "ext_test.variadic2"(%c1_i32, %c1_i32, %c1_i32) {operandSegmentSizes = array<i32: 3>, resultSegmentSizes = array<i32: 0>} : (i32, i32, i32) -> ()
            v5 = Variadic2Op([], [ione, ione, ione])
            # CHECK: %7:2 = "ext_test.variadic2"(%c1_i32) {operandSegmentSizes = array<i32: 1>, resultSegmentSizes = array<i32: 2>} : (i32) -> (i32, i32)
            v6 = Variadic2Op([i32, i32], [ione])

            # CHECK: %8 = "ext_test.mixed"(%c1_i32, %c1_i32) {in2 = 2 : i32, in4 = 2 : i32, operandSegmentSizes = array<i32: 1, 1, 0>} : (i32, i32) -> i32
            m1 = MixedOp(i32, ione, iattr, iattr, ione)
            # CHECK: %9 = "ext_test.mixed"(%c1_i32, %c1_i32, %c1_i32) {in2 = 2 : i32, in4 = 2 : i32, operandSegmentSizes = array<i32: 1, 1, 1>} : (i32, i32, i32) -> i32
            m2 = MixedOp(i32, ione, iattr, iattr, ione, in3=ione)

        print(module)
        assert module.operation.verify()

        # CHECK: OpResult(%c1_i32
        print(c1.a)
        # CHECK: 2 : i32
        print(c1.x)
        # CHECK: OpResult(%c1_i32
        print(o1.a)
        # CHECK: None
        print(o1.b)
        # CHECK: OpResult(%c1_i32
        print(o2.b)
        # CHECK: 0
        print(o1.out1.result_number)
        # CHECK: None
        print(o1.out2)
        # CHECK: 0
        print(o2.out1.result_number)
        # CHECK: 2
        print(o2.out2.result_number)
        # CHECK: None
        print(o3.a)
        # CHECK: OpResult(%c1_i32
        print(o5.a)
        # CHECK: ['OpResult(%c1_i32 = arith.constant 1 : i32)', 'OpResult(%c1_i32 = arith.constant 1 : i32)']
        print([str(i) for i in v1.c])
        # CHECK: ['OpResult(%c1_i32 = arith.constant 1 : i32)']
        print([str(i) for i in v2.c])
        # CHECK: []
        print([str(i) for i in v3.c])
        # CHECK: 0 0
        print(len(v4.a), len(v4.b))
        # CHECK: 3 0
        print(len(v5.a), len(v5.b))
        # CHECK: 1 2
        print(len(v6.a), len(v6.b))

        # cases to violate constraits
        module = Module.create()
        with InsertionPoint(module.body):
            try:
                c1 = ConstraintOp(ione, ione, fone, ione, iattr)
            except TypeError as e:
                # CHECK: missing a required argument: 'y'
                print(e)

            try:
                c2 = ConstraintOp(ione, ione, fone, ione, iattr, fattr, ione)
            except TypeError as e:
                # CHECK：too many positional arguments
                print(e)


# CHECK: TEST: testExtDialectWithRegion
@run
def testExtDialectWithRegion():
    class ParentIsIfTrait(DynamicOpTrait):
        @staticmethod
        def verify_invariants(op) -> bool:
            if not isinstance(op.parent.opview, IfOp):
                op.location.emit_error(
                    f"{op.name} should be put inside {IfOp.OPERATION_NAME}"
                )
                return False
            return True

    class TestRegion(Dialect, name="ext_region"):
        pass

    class IfOp(TestRegion.Operation, name="if"):
        result: Result[Any]
        cond: Operand[IntegerType[1]]
        then: Region
        else_: Region

    class YieldOp(
        TestRegion.Operation, name="yield", traits=[IsTerminatorTrait, ParentIsIfTrait]
    ):
        value: Operand[Any]

        def verify_invariants(self) -> bool:
            if self.parent.results[0].type != self.value.type:
                self.location.emit_error(
                    "result type mismatch between YieldOp and its parent IfOp"
                )
                return False
            return True

    class NoTermOp(TestRegion.Operation, name="no_term", traits=[NoTerminatorTrait]):
        body: Region

    with Context(), Location.unknown():
        TestRegion.load()
        # CHECK: irdl.dialect @ext_region {
        # CHECK:   irdl.operation @if {
        # CHECK:     %0 = irdl.is i1
        # CHECK:     irdl.operands(cond: %0)
        # CHECK:     %1 = irdl.any
        # CHECK:     irdl.results(result: %1)
        # CHECK:     %2 = irdl.region
        # CHECK:     %3 = irdl.region
        # CHECK:     irdl.regions(then: %2, else_: %3)
        # CHECK:   }
        # CHECK:   irdl.operation @yield {
        # CHECK:     %0 = irdl.any
        # CHECK:     irdl.operands(value: %0)
        # CHECK:   }
        # CHECK:   irdl.operation @no_term {
        # CHECK:     %0 = irdl.region
        # CHECK:     irdl.regions(body: %0)
        # CHECK:   }
        # CHECK: }
        print(TestRegion._mlir_module)

        # CHECK: (self, /, result, cond, *, loc=None, ip=None)
        print(IfOp.__init__.__signature__)

        # CHECK: None None
        print(IfOp._ODS_OPERAND_SEGMENTS, IfOp._ODS_RESULT_SEGMENTS)
        # CHECK: (2, True)
        print(IfOp._ODS_REGIONS)

        module = Module.create()
        with InsertionPoint(module.body):
            i1 = IntegerType.get_signless(1)
            i32 = IntegerType.get_signless(32)
            cond = arith.constant(i1, 1)

            if_ = IfOp(i32, cond)
            if_.then.blocks.append()
            if_.else_.blocks.append()

            with InsertionPoint(if_.then.blocks[0]):
                v = arith.constant(i32, 2)
                yield_ = YieldOp(v)

            with InsertionPoint(if_.else_.blocks[0]):
                v = arith.constant(i32, 3)
                YieldOp(v)

            nt = NoTermOp()
            nt.body.blocks.append()

            with InsertionPoint(nt.body.blocks[0]):
                arith.constant(i32, 4)
                # No terminator here

        assert module.operation.verify()
        # CHECK: module {
        # CHECK:     %true = arith.constant true
        # CHECK:     %0 = "ext_region.if"(%true) ({
        # CHECK:         %c2_i32 = arith.constant 2 : i32
        # CHECK:         "ext_region.yield"(%c2_i32) : (i32) -> ()
        # CHECK:     }, {
        # CHECK:         %c3_i32 = arith.constant 3 : i32
        # CHECK:         "ext_region.yield"(%c3_i32) : (i32) -> ()
        # CHECK:     }) : (i1) -> i32
        # CHECK:     "ext_region.no_term"() ({
        # CHECK:       %c4_i32 = arith.constant 4 : i32
        # CHECK:     }) : () -> ()
        # CHECK: }
        print(module)

        # CHECK: True
        print(yield_.has_trait(IsTerminatorTrait))
        # CHECK: False
        print(yield_.has_trait(NoTerminatorTrait))
        # CHECK: True
        print(yield_.has_trait(ParentIsIfTrait))
        # CHECK: False
        print(nt.operation.has_trait(IsTerminatorTrait))
        # CHECK: True
        print(nt.operation.has_trait(NoTerminatorTrait))
        # CHECK: False
        print(nt.operation.has_trait(ParentIsIfTrait))
        # CHECK: False
        print(NoTermOp.has_trait(IsTerminatorTrait))
        # CHECK: True
        print(NoTermOp.has_trait(NoTerminatorTrait))
        # CHECK: False
        print(NoTermOp.has_trait(ParentIsIfTrait))

        # CHECK: %c2_i32 = arith.constant 2 : i32
        print(if_.then.blocks[0])
        # CHECK: %c3_i32 = arith.constant 3 : i32
        print(if_.else_.blocks[0])

        # CHECK-LABEL: Testing violation cases
        print("Testing violation cases:")

        module = Module.create()
        with InsertionPoint(module.body):
            i1 = IntegerType.get_signless(1)
            i32 = IntegerType.get_signless(32)
            cond = arith.constant(i1, 1)

            if_ = IfOp(i32, cond)
            if_.then.blocks.append()
            if_.else_.blocks.append()

            with InsertionPoint(if_.then.blocks[0]):
                v = arith.constant(i32, 2)

            with InsertionPoint(if_.else_.blocks[0]):
                v = arith.constant(i32, 3)

        try:
            module.operation.verify()
        except Exception as e:
            # CHECK: Verification failed:
            # CHECK: block with no terminator
            print(e)

        module = Module.create()
        with InsertionPoint(module.body):
            v = arith.constant(i32, 2)
            YieldOp(v)

        try:
            module.operation.verify()
        except Exception as e:
            # CHECK: Verification failed:
            # CHECK: ext_region.yield should be put inside ext_region.if
            print(e)

        module = Module.create()
        with InsertionPoint(module.body):
            i1 = IntegerType.get_signless(1)
            i32 = IntegerType.get_signless(32)
            cond = arith.constant(i1, 1)

            if_ = IfOp(i1, cond)
            if_.then.blocks.append()
            if_.else_.blocks.append()

            with InsertionPoint(if_.then.blocks[0]):
                v = arith.constant(i32, 2)
                YieldOp(v)

            with InsertionPoint(if_.else_.blocks[0]):
                v = arith.constant(i32, 3)
                YieldOp(v)

        try:
            module.operation.verify()
        except Exception as e:
            # CHECK: Verification failed:
            # CHECK: result type mismatch
            print(e)


# CHECK: TEST: testExtDialectWithType
@run
def testExtDialectWithType():
    class TestType(Dialect, name="ext_type"):
        pass

    class Array(TestType.Type, name="array"):
        elem_type: IntegerType[32] | IntegerType[64]
        length: IntegerAttr

    class MakeArrayOp(TestType.Operation, name="make_array"):
        arr: Result[Array]

    class MakeArray3Op(TestType.Operation, name="make_array3"):
        arr: Result[Array[IntegerType[32], IntegerAttr[IntegerType[32], 3]]] = result(
            infer_type=True
        )

    with Context(), Location.unknown():
        TestType.load()
        # CHECK: irdl.dialect @ext_type {
        # CHECK:   irdl.type @array {
        # CHECK:     %0 = irdl.is i32
        # CHECK:     %1 = irdl.is i64
        # CHECK:     %2 = irdl.any_of(%0, %1)
        # CHECK:     %3 = irdl.base "#builtin.integer"
        # CHECK:     irdl.parameters(elem_type: %2, length: %3)
        # CHECK:   }
        # CHECK:   irdl.operation @make_array {
        # CHECK:     %0 = irdl.base @ext_type::@array
        # CHECK:     irdl.results(arr: %0)
        # CHECK:   }
        # CHECK:   irdl.operation @make_array3 {
        # CHECK:     %0 = irdl.is i32
        # CHECK:     %1 = irdl.is 3 : i32
        # CHECK:     %2 = irdl.parametric @ext_type::@array<%0, %1>
        # CHECK:     irdl.results(arr: %2)
        # CHECK:   }
        # CHECK: }
        print(TestType._mlir_module)

        # CHECK: ext_type.array
        print(Array.type_name)

        i32 = IntegerType.get_signless(32)
        i64 = IntegerType.get_signless(64)
        a4 = Array.get(i32, IntegerAttr.get(i32, 4))
        a6 = Array.get(i64, IntegerAttr.get(i32, 6))
        # CHECK: !ext_type.array<i32, 4 : i32>
        print(a4)
        # CHECK: !ext_type.array<i64, 6 : i32>
        print(a6)

        # CHECK: i32
        print(a4.elem_type)
        # CHECK: 4 : i32
        print(a4.length)
        # CHECK: i64
        print(a6.elem_type)
        # CHECK: 6 : i32
        print(a6.length)

        # CHECK: <locals>.Array
        print(type(Type(a4).maybe_downcast()))

        module = Module.create()
        with InsertionPoint(module.body):
            MakeArrayOp(a4)
            MakeArrayOp(a6)
            MakeArray3Op()

        # CHECK: %0 = "ext_type.make_array"() : () -> !ext_type.array<i32, 4 : i32>
        # CHECK: %1 = "ext_type.make_array"() : () -> !ext_type.array<i64, 6 : i32>
        # CHECK: %2 = "ext_type.make_array3"() : () -> !ext_type.array<i32, 3 : i32>
        assert module.operation.verify()
        print(module)


# CHECK: TEST: testExtDialectWithAttr
@run
def testExtDialectWithAttr():
    class TestAttr(Dialect, name="ext_attr"):
        pass

    class IntPair(TestAttr.Attribute, name="pair"):
        first: IntegerAttr
        second: IntegerAttr

    class StrPair(TestAttr.Attribute, name="str_pair"):
        first: StringAttr
        second: StringAttr

    class Op1(TestAttr.Operation, name="op1"):
        pair: IntPair

    class Op2(TestAttr.Operation, name="op2"):
        pair: StrPair
        pair2: StrPair[StringAttr["a"], StringAttr["b"]]

    with Context(), Location.unknown():
        TestAttr.load()
        # CHECK: irdl.dialect @ext_attr {
        # CHECK:   irdl.attribute @pair {
        # CHECK:     %0 = irdl.base "#builtin.integer"
        # CHECK:     %1 = irdl.base "#builtin.integer"
        # CHECK:     irdl.parameters(first: %0, second: %1)
        # CHECK:   }
        # CHECK:   irdl.attribute @str_pair {
        # CHECK:     %0 = irdl.base "#builtin.string"
        # CHECK:     %1 = irdl.base "#builtin.string"
        # CHECK:     irdl.parameters(first: %0, second: %1)
        # CHECK:   }
        # CHECK:   irdl.operation @op1 {
        # CHECK:     %0 = irdl.base @ext_attr::@pair
        # CHECK:     irdl.attributes {"pair" = %0}
        # CHECK:   }
        # CHECK:   irdl.operation @op2 {
        # CHECK:     %0 = irdl.base @ext_attr::@str_pair
        # CHECK:     %1 = irdl.is "a"
        # CHECK:     %2 = irdl.is "b"
        # CHECK:     %3 = irdl.parametric @ext_attr::@str_pair<%1, %2>
        # CHECK:     irdl.attributes {"pair" = %0, "pair2" = %3}
        # CHECK:   }
        # CHECK: }
        print(TestAttr._mlir_module)

        # CHECK: ext_attr.pair
        print(IntPair.attr_name)

        # CHECK: ext_attr.str_pair
        print(StrPair.attr_name)

        ip = IntPair.get(
            IntegerAttr.get(IntegerType.get_signless(32), 1),
            IntegerAttr.get(IntegerType.get_signless(32), 2),
        )
        sp = StrPair.get(StringAttr.get("hello"), StringAttr.get("world"))
        # CHECK: #ext_attr.pair<1 : i32, 2 : i32>
        print(ip)
        # CHECK: #ext_attr.str_pair<"hello", "world">
        print(sp)

        # CHECK: "hello"
        print(sp.first)
        # CHECK: "world"
        print(sp.second)

        sp2 = ir.Attribute(sp).maybe_downcast()
        # CHECK: <locals>.StrPair
        print(type(sp2))
        # CHECK: #ext_attr.str_pair<"hello", "world">
        print(str(sp2))

        module = Module.create()
        with InsertionPoint(module.body):
            Op1(ip)
            p2 = StrPair.get(StringAttr.get("a"), StringAttr.get("b"))
            Op2(sp, p2)

        assert module.operation.verify()

        # CHECK: "ext_attr.op1"() {pair = #ext_attr.pair<1 : i32, 2 : i32>} : () -> ()
        # CHECK: "ext_attr.op2"() {pair = #ext_attr.str_pair<"hello", "world">, pair2 = #ext_attr.str_pair<"a", "b">} : () -> ()
        print(module)


# CHECK: TEST: testExtDialectWithInvalidOp
@run
def testExtDialectWithInvalidOp():
    class TestInvalid(Dialect, name="ext_invalid"):
        pass

    try:

        class InferTypeBeforePositionalOp(
            TestInvalid.Operation, name="infer_before_pos"
        ):
            res: Result[IntegerType[32]] = result(infer_type=True)
            a: Operand[IntegerType[32]]

    except ValueError as e:
        # CHECK: wrong parameter order
        print(e)

    try:

        class AssignNoneOnNonOptionalOp(
            TestInvalid.Operation, name="assign_none_on_non_optional"
        ):
            a: Operand[IntegerType[32]] = None

    except ValueError as e:
        # CHECK: only optional operand can be set to None
        print(e)

    try:

        class AssignNoneOnnAttributeOp(
            TestInvalid.Operation, name="assign_none_on_attribute"
        ):
            a: IntegerAttr = None

    except ValueError as e:
        # CHECK: only optional attribute can be set to None
        print(e)

    try:

        class CannotInferTypeOp(TestInvalid.Operation, name="cannot_infer_type"):
            a: Result[IntegerType] = result(infer_type=True)

    except TypeError as e:
        # CHECK: unsupported type for inferring
        print(e)

    try:

        class WrongFieldSpecifierOp(
            TestInvalid.Operation, name="wrong_field_specifier"
        ):
            a: Result[IntegerType] = operand()

    except TypeError as e:
        # CHECK: only `result` field specifier can be used for result fields
        print(e)

    try:

        class WrongFieldSpecifierOp2(
            TestInvalid.Operation, name="wrong_field_specifier2"
        ):
            a: IntegerAttr = operand()

    except TypeError as e:
        # CHECK: only `attribute` field specifier can be used for attribute fields
        print(e)


# CHECK: TEST: testExtDialectWithAttrInOp
@run
def testExtDialectWithAttrInOp():
    class TestAttrInOp(Dialect, name="ext_attr_in_op"):
        pass

    class OpWithAttr(TestAttrInOp.Operation, name="op_with_attr"):
        a: IntegerAttr | StringAttr
        b: IntegerType[32] | IntegerType[64]

    with Context(), Location.unknown():
        TestAttrInOp.load()
        # CHECK: irdl.dialect @ext_attr_in_op {
        # CHECK:   irdl.operation @op_with_attr {
        # CHECK:     %0 = irdl.base "#builtin.integer"
        # CHECK:     %1 = irdl.base "#builtin.string"
        # CHECK:     %2 = irdl.any_of(%0, %1)
        # CHECK:     %3 = irdl.is i32
        # CHECK:     %4 = irdl.is i64
        # CHECK:     %5 = irdl.any_of(%3, %4)
        # CHECK:     irdl.attributes {"a" = %2, "b" = %5}
        # CHECK:   }
        # CHECK: }
        print(TestAttrInOp._mlir_module)

        i32 = IntegerType.get_signless(32)
        i64 = IntegerType.get_signless(64)
        iattr = IntegerAttr.get(i32, 42)
        sattr = StringAttr.get("hello")

        module = Module.create()
        with InsertionPoint(module.body):
            OpWithAttr(iattr, TypeAttr.get(i32))
            OpWithAttr(sattr, TypeAttr.get(i64))

        assert module.operation.verify()
        # CHECK: "ext_attr_in_op.op_with_attr"() {a = 42 : i32, b = i32} : () -> ()
        # CHECK: "ext_attr_in_op.op_with_attr"() {a = "hello", b = i64} : () -> ()
        print(module)


@run
def testExtDialectFieldSpecifiers():
    class TestFieldSpecifiers(Dialect, name="ext_field_specifiers"):
        pass

    class OperandSpecifierOp(TestFieldSpecifiers.Operation, name="operand_specifier"):
        a: Operand[IntegerType[32]] = operand()
        b: Optional[Operand[IntegerType[32]]] = None
        c: Operand[IntegerType[32]] = operand(kw_only=True)

    class ResultSpecifierOp(TestFieldSpecifiers.Operation, name="result_specifier"):
        a: Result[IntegerType[32]] = result()
        b: Result[IntegerType[16]] = result(infer_type=True)
        c: Result[IntegerType] = result(
            default_factory=lambda: IntegerType.get_signless(8)
        )
        d: Sequence[Result[IntegerType]] = result(default_factory=list)
        e: Result[IntegerType[32]] = result(kw_only=True)

    class AttributeSpecifierOp(
        TestFieldSpecifiers.Operation, name="attribute_specifier"
    ):
        a: IntegerAttr = attribute()
        b: IntegerAttr = attribute(
            default_factory=lambda: IntegerAttr.get(IntegerType.get_signless(32), 42)
        )
        c: StringAttr["a"] | StringAttr["b"] = attribute(
            default_factory=lambda: StringAttr.get("a"), kw_only=True
        )
        d: IntegerAttr = attribute(kw_only=True)

    with Context(), Location.unknown():
        TestFieldSpecifiers.load()

        # CHECK: (self, /, a, *, b=None, c, loc=None, ip=None)
        print(OperandSpecifierOp.__init__.__signature__)
        # CHECK: (self, /, a, *, b=None, c=None, d=None, e, loc=None, ip=None)
        print(ResultSpecifierOp.__init__.__signature__)
        # CHECK: (self, /, a, *, b=None, c=None, d, loc=None, ip=None)
        print(AttributeSpecifierOp.__init__.__signature__)

        module = Module.create()
        i32 = IntegerType.get_signless(32)
        with InsertionPoint(module.body):
            one = arith.constant(i32, 1)
            OperandSpecifierOp(one, c=one)
            ResultSpecifierOp(i32, e=i32)
            AttributeSpecifierOp(IntegerAttr.get(i32, 43), d=IntegerAttr.get(i32, 100))

        assert module.operation.verify()
        # CHECK: "ext_field_specifiers.operand_specifier"(%c1_i32, %c1_i32) {operandSegmentSizes = array<i32: 1, 0, 1>} : (i32, i32) -> ()
        # CHECK: %0:4 = "ext_field_specifiers.result_specifier"() {resultSegmentSizes = array<i32: 1, 1, 1, 0, 1>} : () -> (i32, i16, i8, i32)
        # CHECK: "ext_field_specifiers.attribute_specifier"() {a = 43 : i32, b = 42 : i32, c = "a", d = 100 : i32} : () -> ()
        print(module)



"""Test breaking inside functions defined within a BSD archive file libfoo.a."""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil
import os
import time


class BSDArchivesTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False
    # If your test case doesn't stress debug info, then
    # set this to true.  That way it won't be run once for
    # each debug info format.
    NO_DEBUG_INFO_TESTCASE = True

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Find the line number in a(int) to break at.
        self.line = line_number("a.c", "// Set file and line breakpoint inside a().")

    def test(self):
        """Break inside a() and b() defined within libfoo.a."""
        self.build()

        exe = self.getBuildArtifact("a.out")
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Break inside a() by file and line first.
        lldbutil.run_break_set_by_file_and_line(
            self, "a.c", self.line, num_expected_locations=1, loc_exact=True
        )

        self.runCmd("run", RUN_SUCCEEDED)

        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Break at a(int) first.
        self.expect(
            "frame variable", VARIABLES_DISPLAYED_CORRECTLY, substrs=["(int) arg = 1"]
        )
        self.expect_var_path("__a_global", type="int", value="1")

        # Set breakpoint for b() next.
        lldbutil.run_break_set_by_symbol(
            self, "b", num_expected_locations=1, sym_exact=True
        )

        # Continue the program, we should break at b(int) next.
        self.runCmd("continue")
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )
        self.expect(
            "frame variable", VARIABLES_DISPLAYED_CORRECTLY, substrs=["(int) arg = 2"]
        )
        self.expect_var_path("__b_global", type="int", value="2")

    def check_frame_variable_errors(self, thread, error_strings):
        command_result = lldb.SBCommandReturnObject()
        interp = self.dbg.GetCommandInterpreter()
        result = interp.HandleCommand("frame variable", command_result)
        self.assertEqual(
            result, lldb.eReturnStatusFailed, "frame var succeeded unexpectedly"
        )
        command_error = command_result.GetError()

        frame = thread.GetFrameAtIndex(0)
        var_list = frame.GetVariables(True, True, False, True)
        self.assertEqual(var_list.GetSize(), 0)
        api_error = var_list.GetError().GetCString()

        for s in error_strings:
            self.assertIn(
                s,
                command_error,
                'Make sure "%s" exists in the command error "%s"' % (s, command_error),
            )
        for s in error_strings:
            self.assertIn(
                s,
                api_error,
                'Make sure "%s" exists in the API error "%s"' % (s, api_error),
            )

    @skipIfRemote
    @skipUnlessDarwin
    def test_frame_var_errors_when_archive_missing(self):
        """
        Break inside a() and remove libfoo.a to make sure we can't load
        the debug information and report an appropriate error when doing
        'frame variable'.
        """
        self.build()
        exe = self.getBuildArtifact("a.out")
        libfoo_path = self.getBuildArtifact("libfoo.a")
        # Delete the main.o file that contains the debug info so we force an
        # error when we run to main and try to get variables for the a()
        # function. Since the libfoo.a is missing, the debug info won't be
        # loaded and we should see an error when trying to read varibles.
        os.unlink(libfoo_path)

        (target, process, thread, bkpt) = lldbutil.run_to_name_breakpoint(
            self, "a", bkpt_module=exe
        )

        error_strings = [
            'debug map object file "',
            'libfoo.a(a.o)" containing debug info does not exist, debug info will not be loaded',
        ]
        self.check_frame_variable_errors(thread, error_strings)

    @skipIfRemote
    @skipIf(compiler="clang", compiler_version=["<", "12.0"])
    def test_archive_specifications(self):
        """
        Create archives and make sure the information we get when retrieving
        the modules specifications is correct.
        """
        self.build()
        libbar_path = self.getBuildArtifact("libbar.a")
        libfoo_path = self.getBuildArtifact("libfoo.a")
        libfoothin_path = self.getBuildArtifact("libfoo-thin.a")
        objfile_a = self.getBuildArtifact("a.o")
        objfile_b = self.getBuildArtifact("b.o")
        objfile_c = self.getBuildArtifact("c.o")
        size_a = os.path.getsize(objfile_a)
        size_b = os.path.getsize(objfile_b)
        size_c = os.path.getsize(objfile_c)

        # Test loading normal archives
        module_specs = lldb.SBModuleSpecList.GetModuleSpecifications(libfoo_path)
        num_specs = module_specs.GetSize()
        self.assertEqual(num_specs, 2)
        spec = module_specs.GetSpecAtIndex(0)
        self.assertEqual(spec.GetObjectName(), "a.o")
        self.assertEqual(spec.GetObjectSize(), size_a)
        spec = module_specs.GetSpecAtIndex(1)
        self.assertEqual(spec.GetObjectName(), "b.o")
        self.assertEqual(spec.GetObjectSize(), size_b)

        # Test loading thin archives
        module_specs = lldb.SBModuleSpecList.GetModuleSpecifications(libbar_path)
        num_specs = module_specs.GetSize()
        self.assertEqual(num_specs, 1)
        spec = module_specs.GetSpecAtIndex(0)
        self.assertEqual(spec.GetObjectName(), "c.o")
        self.assertEqual(spec.GetObjectSize(), size_c)

        module_specs = lldb.SBModuleSpecList.GetModuleSpecifications(libfoothin_path)
        num_specs = module_specs.GetSize()
        self.assertEqual(num_specs, 2)
        spec = module_specs.GetSpecAtIndex(0)
        self.assertEqual(spec.GetObjectName(), "a.o")
        self.assertEqual(spec.GetObjectSize(), size_a)
        spec = module_specs.GetSpecAtIndex(1)
        self.assertEqual(spec.GetObjectName(), "b.o")
        self.assertEqual(spec.GetObjectSize(), size_b, libfoothin_path)

    @skipIfRemote
    @skipUnlessDarwin
    def test_frame_var_errors_when_thin_archive_malformed(self):
        """
        Create thin archive libfoo.a and make it malformed to make sure
        we don't crash and report an appropriate error when resolving
        breakpoint using debug map.
        """
        self.build()
        exe = self.getBuildArtifact("a.out")
        libfoo_path = self.getBuildArtifact("libfoo.a")
        libthin_path = self.getBuildArtifact("libfoo-thin.a")
        objfile_a = self.getBuildArtifact("a.o")
        objfile_b = self.getBuildArtifact("b.o")
        objfile_c = self.getBuildArtifact("c.o")
        # Replace the libfoo.a file with a thin archive containing the same
        # debug information (a.o, b.o). Then remove a.o from the file system
        # so we force an error when we set a breakpoint on a() function.
        # Since the a.o is missing, the debug info won't be loaded and we
        # should see an error when trying to break into a().
        os.remove(libfoo_path)
        shutil.copyfile(libthin_path, libfoo_path)
        os.remove(objfile_a)

        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)
        # We won't be able to see source file
        self.expect(
            "b a",
            substrs=["Breakpoint 1: where = a.out`a, address ="],
        )
        # Break at a() should fail
        (target, process, thread, bkpt) = lldbutil.run_to_name_breakpoint(
            self, "a", bkpt_module=exe
        )
        error_strings = [
            '"a.o" object from the "',
            "libfoo.a\" archive: either the .o file doesn't exist in the archive or the modification time (0x",
            ") of the .o file doesn't match",
        ]
        self.check_frame_variable_errors(thread, error_strings)

        # Break at b() should succeed
        (target, process, thread, bkpt) = lldbutil.run_to_name_breakpoint(
            self, "b", bkpt_module=exe
        )
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )
        self.expect(
            "frame variable", VARIABLES_DISPLAYED_CORRECTLY, substrs=["(int) arg = 2"]
        )

    @skipIfRemote
    @skipUnlessDarwin
    def test_frame_var_errors_when_mtime_mistmatch_for_object_in_archive(self):
        """
        Break inside a() and modify the modification time for "a.o" within
        libfoo.a to make sure we can't load the debug information and
        report an appropriate error when doing 'frame variable'.
        """
        self.build()
        exe = self.getBuildArtifact("a.out")
        a_path = self.getBuildArtifact("a.o")

        # Change the modification time of the a.o object file after sleeping for
        # 2 seconds to ensure the modification time is different. The rebuild
        # only the "libfoo.a" target. This means the modification time of the
        # a.o within libfoo.a will not match the debug map's modification time
        # in a.out and will cause the debug information to not be loaded and we
        # should get an appropriate error when reading variables.
        time.sleep(2)
        os.utime(a_path, None)
        self.build(make_targets=["libfoo.a"])

        (target, process, thread, bkpt) = lldbutil.run_to_name_breakpoint(
            self, "a", bkpt_module=exe
        )

        error_strings = [
            '"a.o" object from the "',
            "libfoo.a\" archive: either the .o file doesn't exist in the archive or the modification time (0x",
            ") of the .o file doesn't match",
        ]
        self.check_frame_variable_errors(thread, error_strings)



import lldb

from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil
from lldbgdbserverutils import get_debugserver_exe

import os
import platform
import shutil
import time
import socket


class TestStopAtEntry(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    # The port used by debugserver.
    PORT = 54638

    # The number of attempts.
    ATTEMPTS = 10

    # Time given to the binary to launch and to debugserver to attach to it for
    # every attempt. We'll wait a maximum of 10 times 2 seconds while the
    # inferior will wait 10 times 10 seconds.
    TIMEOUT = 2

    def no_debugserver(self):
        if get_debugserver_exe() is None:
            return "no debugserver"
        return None

    def port_not_available(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if s.connect_ex(("127.0.0.1", self.PORT)) == 0:
            return "{} not available".format(self.PORT)
        return None

    @skipUnlessDarwin
    @skipIfRemote
    def test_stop_default_platform_sync(self):
        self.do_test_stop_at_entry(True, False)

    @skipUnlessDarwin
    @skipIfRemote
    def test_stop_default_platform_async(self):
        self.do_test_stop_at_entry(False, False)

    @skipUnlessDarwin
    @skipIfRemote
    @skipTestIfFn(no_debugserver)
    @skipTestIfFn(port_not_available)
    def test_stop_remote_platform_sync(self):
        self.do_test_stop_at_entry(True, True)

    @skipUnlessDarwin
    @skipIfRemote
    @skipTestIfFn(no_debugserver)
    @skipTestIfFn(port_not_available)
    def test_stop_remote_platform_async(self):
        self.do_test_stop_at_entry(False, True)

    def do_test_stop_at_entry(self, synchronous, remote):
        """Test the normal launch path in either sync or async mode"""
        self.build()

        target = lldbutil.run_to_breakpoint_make_target(self)
        launch_info = target.GetLaunchInfo()
        launch_info.SetLaunchFlags(lldb.eLaunchFlagStopAtEntry)
        old_async = self.dbg.GetAsync()

        def cleanup():
            self.dbg.SetAsync(old_async)

        self.addTearDownHook(cleanup)

        if not synchronous:
            self.dbg.SetAsync(True)
            listener = lldb.SBListener("test-process-listener")
            mask = listener.StartListeningForEventClass(
                self.dbg,
                lldb.SBProcess.GetBroadcasterClassName(),
                lldb.SBProcess.eBroadcastBitStateChanged,
            )
            self.assertEqual(
                mask,
                lldb.SBProcess.eBroadcastBitStateChanged,
                "Got right mask for listener",
            )
            launch_info.SetListener(listener)
        else:
            self.dbg.SetAsync(False)

        if remote:
            self.setup_remote_platform()

        error = lldb.SBError()

        process = target.Launch(launch_info, error)
        self.assertSuccess(error, "Launch failed")
        # If we are asynchronous, we have to wait for the events:
        if not synchronous:
            listener = launch_info.GetListener()
            event = lldb.SBEvent()
            result = listener.WaitForEvent(30, event)
            self.assertTrue(result, "Timed out waiting for event from process")
            state = lldb.SBProcess.GetStateFromEvent(event)
            self.assertState(
                state, lldb.eStateStopped, "Didn't get a stopped state after launch"
            )

        # Okay, we should be stopped.  Make sure we are indeed at the
        # entry point.  I only know how to do this on darwin:
        self.assertEqual(
            len(process.threads), 1, "Should only have one thread at entry"
        )
        thread = process.threads[0]
        frame = thread.GetFrameAtIndex(0)
        stop_func = frame.name
        self.assertEqual(stop_func, "_dyld_start")

        # Now make sure that we can resume the process and have it exit.
        error = process.Continue()
        self.assertSuccess(error, "Error continuing")
        # Fetch events till we get eStateExited:
        if not synchronous:
            # Get events till exited.
            listener = launch_info.GetListener()
            event = lldb.SBEvent()
            # We get two running events in a row here???  That's a bug
            # but not the one I'm testing for, so for now just fetch as
            # many as were sent.
            num_running = 0
            state = lldb.eStateRunning
            while state == lldb.eStateRunning:
                num_running += 1
                result = listener.WaitForEvent(30, event)
                self.assertTrue(result, "Timed out waiting for running")
                state = lldb.SBProcess.GetStateFromEvent(event)
                if num_running == 1:
                    self.assertState(state, lldb.eStateRunning, "Got running event")
            # The last event we should get is the exited event
            self.assertState(state, lldb.eStateExited, "Got exit event")
        else:
            # Make sure that the process has indeed exited
            state = process.GetState()
            self.assertState(state, lldb.eStateExited)

    def setup_remote_platform(self):
        return
        self.build()

        exe = self.getBuildArtifact("a.out")
        # Launch our test binary.

        # Attach to it with debugserver.
        debugserver = get_debugserver_exe()
        debugserver_args = ["localhost:{}".format(self.PORT)]
        self.spawnSubprocess(debugserver, debugserver_args)

        # Select the platform.
        self.expect("platform select remote-macosx", substrs=[sdk_dir])

        # Connect to debugserver
        interpreter = self.dbg.GetCommandInterpreter()
        connected = False
        for i in range(self.ATTEMPTS):
            result = lldb.SBCommandReturnObject()
            interpreter.HandleCommand("gdb-remote {}".format(self.PORT), result)
            connected = result.Succeeded()
            if connected:
                break
            time.sleep(self.TIMEOUT)

        self.assertTrue(connected, "could not connect to debugserver")



"""
LLDB AppKit formatters

Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
See https://llvm.org/LICENSE.txt for license information.
SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""
# summary provider for class NSException
import lldb.runtime.objc.objc_runtime
import lldb.formatters.metrics
import CFString
import lldb
import lldb.formatters.Logger

statistics = lldb.formatters.metrics.Metrics()
statistics.add_metric("invalid_isa")
statistics.add_metric("invalid_pointer")
statistics.add_metric("unknown_class")
statistics.add_metric("code_notrun")


class NSKnownException_SummaryProvider:
    def adjust_for_architecture(self):
        pass

    def __init__(self, valobj, params):
        logger = lldb.formatters.Logger.Logger()
        self.valobj = valobj
        self.sys_params = params
        if not (self.sys_params.types_cache.id):
            self.sys_params.types_cache.id = self.valobj.GetType().GetBasicType(
                lldb.eBasicTypeObjCID
            )
        self.update()

    def update(self):
        logger = lldb.formatters.Logger.Logger()
        self.adjust_for_architecture()

    def offset_name(self):
        logger = lldb.formatters.Logger.Logger()
        return self.sys_params.pointer_size

    def offset_reason(self):
        logger = lldb.formatters.Logger.Logger()
        return 2 * self.sys_params.pointer_size

    def description(self):
        logger = lldb.formatters.Logger.Logger()
        name_ptr = self.valobj.CreateChildAtOffset(
            "name", self.offset_name(), self.sys_params.types_cache.id
        )
        reason_ptr = self.valobj.CreateChildAtOffset(
            "reason", self.offset_reason(), self.sys_params.types_cache.id
        )
        return (
            "name:"
            + CFString.CFString_SummaryProvider(name_ptr, None)
            + " reason:"
            + CFString.CFString_SummaryProvider(reason_ptr, None)
        )


class NSUnknownException_SummaryProvider:
    def adjust_for_architecture(self):
        pass

    def __init__(self, valobj, params):
        logger = lldb.formatters.Logger.Logger()
        self.valobj = valobj
        self.sys_params = params
        self.update()

    def update(self):
        logger = lldb.formatters.Logger.Logger()
        self.adjust_for_architecture()

    def description(self):
        logger = lldb.formatters.Logger.Logger()
        stream = lldb.SBStream()
        self.valobj.GetExpressionPath(stream)
        name_vo = self.valobj.CreateValueFromExpression(
            "name", "(NSString*)[" + stream.GetData() + " name]"
        )
        reason_vo = self.valobj.CreateValueFromExpression(
            "reason", "(NSString*)[" + stream.GetData() + " reason]"
        )
        if name_vo.IsValid() and reason_vo.IsValid():
            return (
                CFString.CFString_SummaryProvider(name_vo, None)
                + " "
                + CFString.CFString_SummaryProvider(reason_vo, None)
            )
        return "<variable is not NSException>"


def GetSummary_Impl(valobj):
    logger = lldb.formatters.Logger.Logger()
    global statistics
    (
        class_data,
        wrapper,
    ) = lldb.runtime.objc.objc_runtime.Utilities.prepare_class_detection(
        valobj, statistics
    )
    if wrapper:
        return wrapper

    name_string = class_data.class_name()
    logger >> "class name is: " + str(name_string)

    if name_string == "NSException":
        wrapper = NSKnownException_SummaryProvider(valobj, class_data.sys_params)
        statistics.metric_hit("code_notrun", valobj)
    else:
        wrapper = NSUnknownException_SummaryProvider(valobj, class_data.sys_params)
        statistics.metric_hit(
            "unknown_class", valobj.GetName() + " seen as " + name_string
        )
    return wrapper


def NSException_SummaryProvider(valobj, dict):
    logger = lldb.formatters.Logger.Logger()
    provider = GetSummary_Impl(valobj)
    if provider is not None:
        if isinstance(
            provider, lldb.runtime.objc.objc_runtime.SpecialSituation_Description
        ):
            return provider.message()
        try:
            summary = provider.description()
        except:
            summary = None
        logger >> "got summary " + str(summary)
        if summary is None:
            summary = "<variable is not NSException>"
        return str(summary)
    return "Summary Unavailable"


def __lldb_init_module(debugger, dict):
    debugger.HandleCommand(
        "type summary add -F NSException.NSException_SummaryProvider NSException"
    )



import lldb

# bunch of different kinds of python callables that should
# all work as commands.


def check(debugger, command, context, result, internal_dict):
    if (
        not isinstance(debugger, lldb.SBDebugger)
        or not isinstance(command, str)
        or not isinstance(result, lldb.SBCommandReturnObject)
        or not isinstance(internal_dict, dict)
        or (not context is None and not isinstance(context, lldb.SBExecutionContext))
    ):
        raise Exception()
    result.AppendMessage("All good.")


def vfoobar(*args):
    check(*args)


def v5foobar(debugger, command, context, result, internal_dict, *args):
    check(debugger, command, context, result, internal_dict)


def foobar(debugger, command, context, result, internal_dict):
    check(debugger, command, context, result, internal_dict)


def foobar4(debugger, command, result, internal_dict):
    check(debugger, command, None, result, internal_dict)


class FooBar:
    @staticmethod
    def sfoobar(debugger, command, context, result, internal_dict):
        check(debugger, command, context, result, internal_dict)

    @classmethod
    def cfoobar(cls, debugger, command, context, result, internal_dict):
        check(debugger, command, context, result, internal_dict)

    def ifoobar(self, debugger, command, context, result, internal_dict):
        check(debugger, command, context, result, internal_dict)

    def __call__(self, debugger, command, context, result, internal_dict):
        check(debugger, command, context, result, internal_dict)

    @staticmethod
    def sfoobar4(debugger, command, result, internal_dict):
        check(debugger, command, None, result, internal_dict)

    @classmethod
    def cfoobar4(cls, debugger, command, result, internal_dict):
        check(debugger, command, None, result, internal_dict)

    def ifoobar4(self, debugger, command, result, internal_dict):
        check(debugger, command, None, result, internal_dict)


class FooBar4:
    def __call__(self, debugger, command, result, internal_dict):
        check(debugger, command, None, result, internal_dict)


FooBarObj = FooBar()

FooBar4Obj = FooBar4()



"""
Test completing types using information from other shared libraries.
"""

import os
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class LimitDebugInfoTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False

    def _check_type(self, target, name):
        exe = target.FindModule(lldb.SBFileSpec("a.out"))
        type_ = exe.FindFirstType(name)
        self.trace("type_: %s" % type_)
        self.assertTrue(type_)
        self.assertTrue(type_.IsTypeComplete())
        base = type_.GetDirectBaseClassAtIndex(0).GetType()
        self.trace("base:%s" % base)
        self.assertTrue(base)
        self.assertEqual(base.GetNumberOfFields(), 0)
        self.assertFalse(base.IsTypeComplete())

    def _check_debug_info_is_limited(self, target):
        # Without other shared libraries we should only see the member declared
        # in the derived class. This serves as a sanity check that we are truly
        # building with limited debug info.
        self._check_type(target, "InheritsFromOne")
        self._check_type(target, "InheritsFromTwo")

        # Check that the statistics show that we had incomplete debug info.
        stats = self.get_stats()
        # Find the a.out module info in the stats and verify it has the
        # "debugInfoHadIncompleteTypes" key value pair set to True
        exe_module_found = False
        for module in stats["modules"]:
            if module["path"].endswith("a.out"):
                self.assertTrue(module["debugInfoHadIncompleteTypes"])
                exe_module_found = True
                break
        self.assertTrue(exe_module_found)
        # Verify that "totalModuleCountWithIncompleteTypes" at the top level
        # is greater than zero which shows we had incomplete debug info in a
        # module
        self.assertGreater(stats["totalModuleCountWithIncompleteTypes"], 0)

    def _check_incomplete_frame_variable_output(self):
        # Check that the display of the "frame variable" output identifies the
        # incomplete types. Currently the expression parser will find the real
        # definition for a type when running an expression for any forcefully
        # completed types, but "frame variable" won't. I hope to fix this with
        # a follow up patch, but if we don't find the actual definition we
        # should clearly show this to the user by showing which types were
        # incomplete. So this will test verifies the expected output for such
        # types. We also need to verify the standard "frame variable" output
        # which will inline all of the members on one line, versus the full
        # output from "frame variable --raw" and a few other options.
        # self.expect("frame variable two_as_member", error=True,
        #     substrs=["no member named 'one' in 'InheritsFromOne'"])

        command_expect_pairs = [
            # Test standard "frame variable" output for types to make sure
            # "<incomplete type>" shows up where we expect it to
            [
                "var two_as_member",
                [
                    "(TwoAsMember) ::two_as_member = (two = <incomplete type>, member = 47)"
                ],
            ],
            [
                "var inherits_from_one",
                [
                    "(InheritsFromOne) ::inherits_from_one = (One = <incomplete type>, member = 47)"
                ],
            ],
            [
                "var inherits_from_two",
                [
                    "(InheritsFromTwo) ::inherits_from_two = (Two = <incomplete type>, member = 47)"
                ],
            ],
            [
                "var one_as_member",
                [
                    "(OneAsMember) ::one_as_member = (one = <incomplete type>, member = 47)"
                ],
            ],
            [
                "var two_as_member",
                [
                    "(TwoAsMember) ::two_as_member = (two = <incomplete type>, member = 47)"
                ],
            ],
            [
                "var array_of_one",
                [
                    "(array::One[3]) ::array_of_one = ([0] = <incomplete type>, [1] = <incomplete type>, [2] = <incomplete type>)"
                ],
            ],
            [
                "var array_of_two",
                [
                    "(array::Two[3]) ::array_of_two = ([0] = <incomplete type>, [1] = <incomplete type>, [2] = <incomplete type>)"
                ],
            ],
            [
                "var shadowed_one",
                [
                    "(ShadowedOne) ::shadowed_one = (func_shadow::One = <incomplete type>, member = 47)"
                ],
            ],
            # Now test "frame variable --show-types output" which has multi-line
            # output and should not always show classes that were forcefully
            # completed to the user to let them know they have a type that should
            # have been complete but wasn't.
            [
                "var --show-types inherits_from_one",
                [
                    "(InheritsFromOne) ::inherits_from_one = {",
                    "  (One) One = <incomplete type> {}",
                    "  (int) member = 47",
                    "}",
                ],
            ],
            [
                "var --show-types inherits_from_two",
                [
                    "(InheritsFromTwo) ::inherits_from_two = {",
                    "  (Two) Two = <incomplete type> {}",
                    "  (int) member = 47",
                    "}",
                ],
            ],
            [
                "var  --show-types one_as_member",
                [
                    "(OneAsMember) ::one_as_member = {",
                    "  (member::One) one = <incomplete type> {}",
                    "  (int) member = 47",
                    "}",
                ],
            ],
            [
                "var  --show-types two_as_member",
                [
                    "(TwoAsMember) ::two_as_member = {",
                    "  (member::Two) two = <incomplete type> {}",
                    "  (int) member = 47",
                    "}",
                ],
            ],
            [
                "var  --show-types array_of_one",
                [
                    "(array::One[3]) ::array_of_one = {",
                    "  (array::One) [0] = <incomplete type> {}",
                    "  (array::One) [1] = <incomplete type> {}",
                    "  (array::One) [2] = <incomplete type> {}",
                    "}",
                ],
            ],
            [
                "var  --show-types array_of_two",
                [
                    "(array::Two[3]) ::array_of_two = {",
                    "  (array::Two) [0] = <incomplete type> {}",
                    "  (array::Two) [1] = <incomplete type> {}",
                    "  (array::Two) [2] = <incomplete type> {}",
                    "}",
                ],
            ],
            [
                "var  --show-types shadowed_one",
                [
                    "(ShadowedOne) ::shadowed_one = {",
                    "  (func_shadow::One) func_shadow::One = <incomplete type> {}",
                    "  (int) member = 47",
                    "}",
                ],
            ],
        ]
        for command, expect_items in command_expect_pairs:
            self.expect(command, substrs=expect_items)

    @skipIf(bugnumber="pr46284", debug_info="gmodules")
    @skipIfWindows  # Clang emits type info even with -flimit-debug-info
    # Requires DW_CC_pass_by_* attributes from Clang 7 to correctly call
    # by-value functions.
    @skipIf(compiler="clang", compiler_version=["<", "7.0"])
    def test_one_and_two_debug(self):
        self.build()
        target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))

        self._check_debug_info_is_limited(target)

        lldbutil.run_to_name_breakpoint(self, "main", extra_images=["one", "two"])

        # But when other shared libraries are loaded, we should be able to see
        # all members.
        self.expect_expr("inherits_from_one.member", result_value="47")
        self.expect_expr("inherits_from_one.one", result_value="142")
        self.expect_expr("inherits_from_two.member", result_value="47")
        self.expect_expr("inherits_from_two.one", result_value="142")
        self.expect_expr("inherits_from_two.two", result_value="242")

        self.expect_expr("one_as_member.member", result_value="47")
        self.expect_expr("one_as_member.one.member", result_value="147")
        self.expect_expr("two_as_member.member", result_value="47")
        self.expect_expr("two_as_member.two.one.member", result_value="147")
        self.expect_expr("two_as_member.two.member", result_value="247")

        self.expect_expr("array_of_one[2].member", result_value="174")
        self.expect_expr("array_of_two[2].one[2].member", result_value="174")
        self.expect_expr("array_of_two[2].member", result_value="274")

        self.expect_expr("get_one().member", result_value="124")
        self.expect_expr("get_two().one().member", result_value="124")
        self.expect_expr("get_two().member", result_value="224")

        self.expect_expr("shadowed_one.member", result_value="47")
        self.expect_expr("shadowed_one.one", result_value="142")

        self._check_incomplete_frame_variable_output()

    @skipIf(bugnumber="pr46284", debug_info="gmodules")
    @skipIfWindows  # Clang emits type info even with -flimit-debug-info
    # Requires DW_CC_pass_by_* attributes from Clang 7 to correctly call
    # by-value functions.
    @skipIf(compiler="clang", compiler_version=["<", "7.0"])
    def test_two_debug(self):
        self.build(dictionary=dict(STRIP_ONE="1"))
        target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))

        self._check_debug_info_is_limited(target)

        lldbutil.run_to_name_breakpoint(self, "main", extra_images=["one", "two"])

        # This time, we should only see the members from the second library.
        self.expect_expr("inherits_from_one.member", result_value="47")
        self.expect(
            "expr inherits_from_one.one",
            error=True,
            substrs=["no member named 'one' in 'InheritsFromOne'"],
        )
        self.expect_expr("inherits_from_two.member", result_value="47")
        self.expect(
            "expr inherits_from_two.one",
            error=True,
            substrs=["no member named 'one' in 'InheritsFromTwo'"],
        )
        self.expect_expr("inherits_from_two.two", result_value="242")

        self.expect_expr("one_as_member.member", result_value="47")
        self.expect(
            "expr one_as_member.one.member",
            error=True,
            substrs=["no member named 'member' in 'member::One'"],
        )
        self.expect_expr("two_as_member.member", result_value="47")
        self.expect(
            "expr two_as_member.two.one.member",
            error=True,
            substrs=["no member named 'member' in 'member::One'"],
        )
        self.expect_expr("two_as_member.two.member", result_value="247")

        self.expect(
            "expr array_of_one[2].member",
            error=True,
            substrs=["no member named 'member' in 'array::One'"],
        )
        self.expect(
            "expr array_of_two[2].one[2].member",
            error=True,
            substrs=["no member named 'member' in 'array::One'"],
        )
        self.expect_expr("array_of_two[2].member", result_value="274")

        self.expect(
            "expr get_one().member",
            error=True,
            substrs=["calling 'get_one' with incomplete return type 'result::One'"],
        )
        self.expect(
            "expr get_two().one().member",
            error=True,
            substrs=["calling 'one' with incomplete return type 'result::One'"],
        )
        self.expect_expr("get_two().member", result_value="224")

        self._check_incomplete_frame_variable_output()

    @skipIf(bugnumber="pr46284", debug_info="gmodules")
    @skipIfWindows  # Clang emits type info even with -flimit-debug-info
    # Requires DW_CC_pass_by_* attributes from Clang 7 to correctly call
    # by-value functions.
    @skipIf(compiler="clang", compiler_version=["<", "7.0"])
    def test_one_debug(self):
        self.build(dictionary=dict(STRIP_TWO="1"))
        target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))

        self._check_debug_info_is_limited(target)

        lldbutil.run_to_name_breakpoint(self, "main", extra_images=["one", "two"])

        # In this case we should only see the members from the second library.
        # Note that we cannot see inherits_from_two.one because without debug
        # info for "Two", we cannot determine that it in fact inherits from
        # "One".
        self.expect_expr("inherits_from_one.member", result_value="47")
        self.expect_expr("inherits_from_one.one", result_value="142")
        self.expect_expr("inherits_from_two.member", result_value="47")
        self.expect(
            "expr inherits_from_two.one",
            error=True,
            substrs=["no member named 'one' in 'InheritsFromTwo'"],
        )
        self.expect(
            "expr inherits_from_two.two",
            error=True,
            substrs=["no member named 'two' in 'InheritsFromTwo'"],
        )

        self.expect_expr("one_as_member.member", result_value="47")
        self.expect_expr("one_as_member.one.member", result_value="147")
        self.expect_expr("two_as_member.member", result_value="47")
        self.expect(
            "expr two_as_member.two.one.member",
            error=True,
            substrs=["no member named 'one' in 'member::Two'"],
        )
        self.expect(
            "expr two_as_member.two.member",
            error=True,
            substrs=["no member named 'member' in 'member::Two'"],
        )

        self.expect_expr("array_of_one[2].member", result_value="174")
        self.expect(
            "expr array_of_two[2].one[2].member",
            error=True,
            substrs=["no member named 'one' in 'array::Two'"],
        )
        self.expect(
            "expr array_of_two[2].member",
            error=True,
            substrs=["no member named 'member' in 'array::Two'"],
        )

        self.expect_expr("get_one().member", result_value="124")
        self.expect(
            "expr get_two().one().member",
            error=True,
            substrs=["calling 'get_two' with incomplete return type 'result::Two'"],
        )
        self.expect(
            "expr get_two().member",
            error=True,
            substrs=["calling 'get_two' with incomplete return type 'result::Two'"],
        )

        self._check_incomplete_frame_variable_output()



# RUN: env SUPPORT_LIB=%mlir_cuda_runtime \
# RUN: sh -c 'if [[ "%mlir_run_cuda_sm90_tests" == "1" ]]; \
# RUN: then %PYTHON %s | FileCheck %s; \
# RUN: else export MLIR_NVDSL_PRINT_IR=1; \
# RUN: %PYTHON %s | FileCheck %s --check-prefix=DUMPIR; fi'


# ===----------------------------------------------------------------------===//
#  Chapter 3 : GEMM 128x128x64 with Tensor Core
# ===----------------------------------------------------------------------===//
#
# This program demonstrates a GEMM operation with 128x128x64 matrix multiplication
#
# This chapter introduces demonstrates:
# 1. Execute TMA Load for two input matrices
# 2. Performs Tensor Core GEMM 128x128x64 by warpgroup
# 3. Stores fragmented registers to global memory by warpgroup
#
# ===----------------------------------------------------------------------===//


from mlir import ir
from mlir.dialects import nvgpu, scf, arith, memref, vector, gpu
from tools.nvdsl import *
from mlir.extras import types as T
import numpy as np


def tma_load(
    mbar_group: Mbarriers,
    a_tma: TMA,
    b_tma: TMA,
    p,
):
    """
    TMA loads two input matrices from global memory to shared memory. It performs the following operations:

       - tma.load a_shared_memory[0] at coordinate [0, 0]  (Loads 128x64)
       - tma.load b_shared_memory[0] at coordinate [0, 0]  (Loads 64x64)
       - tma.load b_shared_memory[0] at coordinate [64, 0] (Loads 64x64)

       mbarrier.arrive ta_count = 128x64xf16 + 64x128xf16
    """

    size_tma_a = get_type_size(a_tma.tma_memref)
    size_tma_b = get_type_size(b_tma.tma_memref)
    ta_count = size_tma_a + (size_tma_b * 2)

    off_b = size_tma_a
    off_b2 = off_b + size_tma_b
    a_elem_ty = a_tma.tma_memref.element_type
    b_elem_ty = b_tma.tma_memref.element_type
    a = get_dynamic_shared_memory(a_tma.tma_memref.shape, a_elem_ty)
    b1 = get_dynamic_shared_memory(b_tma.tma_memref.shape, b_elem_ty, off_b)
    b2 = get_dynamic_shared_memory(b_tma.tma_memref.shape, b_elem_ty, off_b2)

    mbar_group[0].arrive(ta_count, predicate=p)

    a_tma.load(a, mbar_group[0], coords=[0, 0], predicate=p)
    b_tma.load(b1, mbar_group[0], coords=[0, 0], predicate=p)
    b_tma.load(b2, mbar_group[0], coords=[64, 0], predicate=p)


@NVDSL.mlir_func
def gemm_128_128_64(a, b, d):
    token_ty = gpu.AsyncTokenType.get()
    t1 = gpu.wait([])
    a_dev, t2 = gpu.alloc(a.type, token_ty, [t1], [], [])
    b_dev, t3 = gpu.alloc(b.type, token_ty, [t2], [], [])
    d_dev, t4 = gpu.alloc(d.type, token_ty, [t3], [], [])
    t5 = gpu.memcpy(token_ty, [t4], a_dev, a)
    t6 = gpu.memcpy(token_ty, [t5], b_dev, b)
    t7 = gpu.wait([t6])

    sw = nvgpu.TensorMapSwizzleKind.SWIZZLE_128B
    a_tma = TMA([128, 64], a.type, swizzle=sw)
    b_tma = TMA([64, 64], b.type, swizzle=sw)
    a_tma.create_descriptor(a_dev)
    b_tma.create_descriptor(b_dev)
    a_size = get_type_size(a.type)
    b_size = get_type_size(b.type)
    smem_size_in_bytes = a_size + b_size

    @NVDSL.mlir_gpu_launch(grid=(1, 1, 1), block=(128, 1, 1), smem=smem_size_in_bytes)
    def gemm_tma_kernel():
        tidx = gpu.thread_id(gpu.Dimension.x)

        mbar_group = Mbarriers(number_of_barriers=1)
        isThread0 = tidx == 0

        mbar_group[0].init(1, predicate=isThread0)
        a_tma.prefetch(predicate=isThread0)
        b_tma.prefetch(predicate=isThread0)

        a_smem = get_dynamic_shared_memory((M, K), T.f16())
        b_smem = get_dynamic_shared_memory((K, N), T.f16(), offset=a_size)

        # 1. TMA Load for two input matrices
        tma_load(mbar_group, a_tma, b_tma, isThread0)

        # 2. All threads wait TMA load completion
        mbar_group[0].try_wait()

        # 3. Performs Tensor Core GEMM 128x128x64 by warpgroup
        A = WGMMAMatrix(WGMMAType.Descriptor, [M, K], desc=a_tma, smem=a_smem)
        B = WGMMAMatrix(WGMMAType.Descriptor, [K, N], desc=b_tma, smem=b_smem)
        D = WGMMAMatrix(WGMMAType.Accumulator, shape=[M, N], ty=T.f32())

        # Matrix Multiply
        D += A @ B

        # 4. Stores fragmented registers to global memory by warpgroup
        D.store_accumulator(d_dev)

    gemm_tma_kernel()

    t8 = gpu.memcpy(token_ty, [t7], d, d_dev)
    gpu.wait([t8])


# Python pass arguments to MLIR
M = 128
N = 128
K = 64
a = np.random.randn(M, K).astype(np.float16)
b = np.random.randn(K, N).astype(np.float16)
d = np.zeros((M, N), np.float32)
gemm_128_128_64(a, b, d)

if os.getenv("MLIR_NVDSL_PRINT_IR") != "1":
    # Verify MLIR program with reference computation in python
    ref_d = a.astype(np.float16) @ b.astype(np.float16)
    np.testing.assert_allclose(d, ref_d, rtol=5e-03, atol=1e-01)
    print("PASS")
# CHECK-NOT: Mismatched elements
# CHECK: PASS

# DUMPIR:   func.func @gemm_128_128_64(%{{.*}}: memref<128x64xf16>, %{{.*}}: memref<64x128xf16>, %[[ARG2:.*]]: memref<128x128xf32>) attributes {llvm.emit_c_interface} {
# DUMPIR:     %[[C128:.*]] = arith.constant 128 : index
# DUMPIR:     %[[C64:.*]] = arith.constant 64 : index
# DUMPIR:     %[[TMA0:.*]] = nvgpu.tma.create.descriptor %{{.*}} box[%[[C128]], %[[C64]]] : memref<*xf16> -> <tensor = memref<128x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>
# DUMPIR:     %[[CAST1:.*]] = memref.cast %{{.*}} : memref<64x128xf16> to memref<*xf16>
# DUMPIR:     %[[C64_5:.*]] = arith.constant 64 : index
# DUMPIR:     %[[C64_6:.*]] = arith.constant 64 : index
# DUMPIR:     %[[TMA1:.*]] = nvgpu.tma.create.descriptor %[[CAST1]] box[%[[C64_5]], %[[C64_6]]] : memref<*xf16> -> <tensor = memref<64x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>
# DUMPIR:       %[[THREADID:.*]] = gpu.thread_id x
# DUMPIR:       %[[MB:.*]] = nvgpu.mbarrier.create -> <memorySpace = #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0:.*]] = arith.constant 0 : index
# DUMPIR:       %[[EQ:.*]] = arith.cmpi eq, %[[THREADID]], %[[C0]] : index
# DUMPIR:       %[[C0_12:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C1_13:.*]] = arith.constant 1 : index
# DUMPIR:       nvgpu.mbarrier.init %[[MB]][%[[C0_12]]], %[[C1_13]], predicate = %[[EQ]] : <memorySpace = #gpu.address_space<workgroup>>
# DUMPIR:       nvgpu.tma.prefetch.descriptor %[[TMA0]], predicate = %[[EQ]] : <tensor = memref<128x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>
# DUMPIR:       nvgpu.tma.prefetch.descriptor %[[TMA1]], predicate = %[[EQ]] : <tensor = memref<64x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>
# DUMPIR:       %[[DSM0:.*]] = gpu.dynamic_shared_memory : memref<?xi8, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_14:.*]] = arith.constant 0 : index
# DUMPIR:       %[[VIEW:.*]] = memref.view %[[DSM0]][%[[C0_14]]][] : memref<?xi8, #gpu.address_space<workgroup>> to memref<128x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[DSM1:.*]] = gpu.dynamic_shared_memory : memref<?xi8, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C16384:.*]] = arith.constant 16384 : index
# DUMPIR:       %[[VIEW_15:.*]] = memref.view %[[DSM1]][%[[C16384]]][] : memref<?xi8, #gpu.address_space<workgroup>> to memref<64x128xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[DSM2:.*]] = gpu.dynamic_shared_memory : memref<?xi8, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_16:.*]] = arith.constant 0 : index
# DUMPIR:       %[[VIEW_17:.*]] = memref.view %[[DSM2]][%[[C0_16]]][] : memref<?xi8, #gpu.address_space<workgroup>> to memref<128x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[DSM3:.*]] = gpu.dynamic_shared_memory : memref<?xi8, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C16384_18:.*]] = arith.constant 16384 : index
# DUMPIR:       %[[VIEW_19:.*]] = memref.view %[[DSM3]][%[[C16384_18]]][] : memref<?xi8, #gpu.address_space<workgroup>> to memref<64x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[DSM4:.*]] = gpu.dynamic_shared_memory : memref<?xi8, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C24576:.*]] = arith.constant 24576 : index
# DUMPIR:       %[[VIEW_20:.*]] = memref.view %[[DSM4]][%[[C24576]]][] : memref<?xi8, #gpu.address_space<workgroup>> to memref<64x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_21:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C32768:.*]] = arith.constant 32768 : index
# DUMPIR:       nvgpu.mbarrier.arrive.expect_tx %[[MB]][%[[C0_21]]], %[[C32768]], predicate = %[[EQ]] : <memorySpace = #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_22:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C0_23:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C0_24:.*]] = arith.constant 0 : index
# DUMPIR:       nvgpu.tma.async.load %[[TMA0]][%[[C0_23]], %[[C0_24]]], %[[MB]][%[[C0_22]]] to %[[VIEW_17]], predicate = %[[EQ]] : <tensor = memref<128x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>, <memorySpace = #gpu.address_space<workgroup>> -> memref<128x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_25:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C0_26:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C0_27:.*]] = arith.constant 0 : index
# DUMPIR:       nvgpu.tma.async.load %[[TMA1]][%[[C0_26]], %[[C0_27]]], %[[MB]][%[[C0_25]]] to %[[VIEW_19]], predicate = %[[EQ]] : <tensor = memref<64x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>, <memorySpace = #gpu.address_space<workgroup>> -> memref<64x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_28:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C64_29:.*]] = arith.constant 64 : index
# DUMPIR:       %[[C0_30:.*]] = arith.constant 0 : index
# DUMPIR:       nvgpu.tma.async.load %[[TMA1]][%[[C64_29]], %[[C0_30]]], %[[MB]][%[[C0_28]]] to %[[VIEW_20]], predicate = %[[EQ]] : <tensor = memref<64x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none>, <memorySpace = #gpu.address_space<workgroup>> -> memref<64x64xf16, #gpu.address_space<workgroup>>
# DUMPIR:       %[[C0_31:.*]] = arith.constant 0 : index
# DUMPIR:       %[[C10000000:.*]] = arith.constant 10000000 : index
# DUMPIR:       %[[FALSE:.*]] = arith.constant false
# DUMPIR:       nvgpu.mbarrier.try_wait.parity %[[MB]][%[[C0_31]]], %[[FALSE]], %[[C10000000]] : <memorySpace = #gpu.address_space<workgroup>>
# DUMPIR:       %[[WG_ACC:.*]] = nvgpu.warpgroup.mma.init.accumulator -> <fragmented = vector<128x128xf32>>
# DUMPIR:       %[[GEN0:.*]] = nvgpu.warpgroup.generate.descriptor %[[VIEW]], %[[TMA0]] : memref<128x64xf16, #gpu.address_space<workgroup>>, <tensor = memref<128x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none> -> <tensor = memref<128x64xf16, #gpu.address_space<workgroup>>>
# DUMPIR:       %[[GEN1:.*]] = nvgpu.warpgroup.generate.descriptor %[[VIEW_15]], %[[TMA1]] : memref<64x128xf16, #gpu.address_space<workgroup>>, <tensor = memref<64x64xf16, 3>, swizzle = swizzle_128b, l2promo = none, oob = zero, interleave = none> -> <tensor = memref<64x128xf16, #gpu.address_space<workgroup>>>
# DUMPIR:       %[[MMA:.*]] = nvgpu.warpgroup.mma %[[GEN0]], %[[GEN1]], %[[WG_ACC]] {transposeB} : <tensor = memref<128x64xf16, #gpu.address_space<workgroup>>>, <tensor = memref<64x128xf16, #gpu.address_space<workgroup>>>, <fragmented = vector<128x128xf32>> -> <fragmented = vector<128x128xf32>>
# DUMPIR:       nvgpu.warpgroup.mma.store %[[MMA]], %{{.*}} : <fragmented = vector<128x128xf32>> to memref<128x128xf32>
# DUMPIR:       gpu.terminator
# DUMPIR:     }
# DUMPIR:     %[[CPY3:.*]] = gpu.memcpy async [%{{.*}}] %[[ARG2]], %{{.*}} : memref<128x128xf32>, memref<128x128xf32>
# DUMPIR:     gpu.wait async [%[[CPY3]]]
# DUMPIR:     return
# DUMPIR:   }



from lldbsuite.test import lldbinline
from lldbsuite.test import decorators

decor = [
    decorators.skipUnlessHasCallSiteInfo,
    decorators.skipIf(dwarf_version=["<", "4"]),
    decorators.skipIf(compiler="clang", compiler_version=["<", "11.0"]),
]
lldbinline.MakeInlineTest(
    __file__,
    globals(),
    name="DisambiguateTailCallSeq_V5",
    build_dict=dict(CFLAGS_EXTRAS="-O2 -glldb"),
    decorators=decor,
)
lldbinline.MakeInlineTest(
    __file__,
    globals(),
    name="DisambiguateTailCallSeq_GNU",
    build_dict=dict(CFLAGS_EXTRAS="-O2 -ggdb"),
    decorators=decor + [decorators.skipIf(debug_info="dsym")],
)



"""
Test SBLineEntry APIs, particularly synthetic line entries.
"""

import os
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class SBLineEntryTestCase(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def test_synthetic_line_entry(self):
        """Test that synthetic LineEntry objects (created via SBAPI) can be
        valid without a valid address range and can be set in SBSymbolContext."""

        # Test creating a synthetic line entry via SBAPI.
        line_entry = lldb.SBLineEntry()
        self.assertFalse(
            line_entry.IsValid(), "Default constructed line entry should be invalid"
        )

        # Set line number - this should mark the line entry as synthetic.
        line_entry.SetLine(42)
        self.assertTrue(
            line_entry.IsValid(),
            "Line entry should be valid after setting line, even without address",
        )
        self.assertEqual(line_entry.GetLine(), 42)

        # Set file and column.
        file_spec = lldb.SBFileSpec(os.path.join(self.getSourceDir(), "test.cpp"), True)
        line_entry.SetFileSpec(file_spec)
        line_entry.SetColumn(10)

        self.assertEqual(line_entry.GetColumn(), 10)
        self.assertEqual(line_entry.GetFileSpec().GetFilename(), "test.cpp")

        # Verify address range is still invalid (synthetic).
        start_addr = line_entry.GetStartAddress()
        self.assertFalse(
            start_addr.IsValid(), "Synthetic line entry should not have valid address"
        )

        # Test setting synthetic line entry in symbol context.
        sym_ctx = lldb.SBSymbolContext()
        sym_ctx.SetLineEntry(line_entry)

        retrieved_line_entry = sym_ctx.GetLineEntry()
        self.assertTrue(
            retrieved_line_entry.IsValid(), "Retrieved line entry should be valid"
        )
        self.assertEqual(retrieved_line_entry.GetLine(), 42)
        self.assertEqual(retrieved_line_entry.GetColumn(), 10)
        self.assertEqual(retrieved_line_entry.GetFileSpec().GetFilename(), "test.cpp")

    def test_line_entry_validity_without_address(self):
        """Test that line entries created via SBAPI are valid without addresses."""

        line_entry = lldb.SBLineEntry()

        # Initially invalid.
        self.assertFalse(line_entry.IsValid())

        # Still invalid with just a file spec.
        file_spec = lldb.SBFileSpec("foo.cpp", True)
        line_entry.SetFileSpec(file_spec)
        self.assertFalse(
            line_entry.IsValid(), "Line entry should be invalid without line number"
        )

        # Valid once line number is set (marks as synthetic).
        line_entry.SetLine(100)
        self.assertTrue(
            line_entry.IsValid(), "Line entry should be valid with line number set"
        )

        # Verify no valid address range.
        self.assertFalse(line_entry.GetStartAddress().IsValid())
        self.assertFalse(line_entry.GetEndAddress().IsValid())

    def test_line_entry_column(self):
        """Test setting and getting column information on synthetic line entries."""

        line_entry = lldb.SBLineEntry()
        line_entry.SetLine(50)

        # Default column should be 0.
        self.assertEqual(line_entry.GetColumn(), 0)

        # Set column.
        line_entry.SetColumn(25)
        self.assertEqual(line_entry.GetColumn(), 25)

        # Verify line entry is still valid.
        self.assertTrue(line_entry.IsValid())

    def test_non_synthetic_line_entry_requires_line_number(self):
        """Test that non-synthetic line entries with addresses still require a line number to be valid."""

        # A line entry is always invalid without a line number, regardless of whether it has an address.
        line_entry = lldb.SBLineEntry()
        self.assertFalse(
            line_entry.IsValid(), "Line entry should be invalid without line number"
        )

        # Even with a file spec, it's still invalid.
        file_spec = lldb.SBFileSpec("test.cpp", True)
        line_entry.SetFileSpec(file_spec)
        self.assertFalse(
            line_entry.IsValid(), "Line entry should be invalid without line number"
        )

        # Only after setting a line number does it become valid.
        line_entry.SetLine(42)
        self.assertTrue(
            line_entry.IsValid(), "Line entry should be valid with line number"
        )

    def test_symbol_context_with_synthetic_line_entry(self):
        """Test that SBSymbolContext correctly stores and retrieves synthetic line entries."""

        # Create a synthetic line entry.
        line_entry = lldb.SBLineEntry()
        line_entry.SetLine(123)
        line_entry.SetColumn(45)
        file_spec = lldb.SBFileSpec("source.cpp", True)
        line_entry.SetFileSpec(file_spec)

        # Create symbol context and set line entry.
        sym_ctx = lldb.SBSymbolContext()
        sym_ctx.SetLineEntry(line_entry)

        # Retrieve and verify.
        retrieved = sym_ctx.GetLineEntry()
        self.assertTrue(retrieved.IsValid())
        self.assertEqual(retrieved.GetLine(), 123)
        self.assertEqual(retrieved.GetColumn(), 45)
        self.assertEqual(retrieved.GetFileSpec().GetFilename(), "source.cpp")

        # Verify it's still synthetic (no valid address).
        self.assertFalse(retrieved.GetStartAddress().IsValid())



"""Test Python APIs for setting, getting, and using address masks."""

import os
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class AddressMasksTestCase(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def reset_all_masks(self, process):
        process.SetAddressMask(
            lldb.eAddressMaskTypeAll,
            lldb.LLDB_INVALID_ADDRESS_MASK,
            lldb.eAddressMaskRangeAll,
        )
        self.runCmd("settings set target.process.virtual-addressable-bits 0")
        self.runCmd("settings set target.process.highmem-virtual-addressable-bits 0")

    @skipIf(archs=["arm$"])  # 32-bit arm ABI hardcodes Code mask, is 32-bit
    def test_address_masks(self):
        self.build()
        (target, process, t, bp) = lldbutil.run_to_source_breakpoint(
            self, "break here", lldb.SBFileSpec("main.c")
        )

        process.SetAddressableBits(lldb.eAddressMaskTypeAll, 42)
        self.assertEqual(0x0000029500003F94, process.FixAddress(0x00265E9500003F94))
        self.reset_all_masks(process)

        # ~((1ULL<<42)-1) == 0xfffffc0000000000
        process.SetAddressMask(lldb.eAddressMaskTypeAll, 0xFFFFFC0000000000)
        self.assertEqual(0x0000029500003F94, process.FixAddress(0x00265E9500003F94))
        self.reset_all_masks(process)

        # Check that all bits can pass through unmodified
        process.SetAddressableBits(lldb.eAddressMaskTypeAll, 64)
        self.assertEqual(0x00265E9500003F94, process.FixAddress(0x00265E9500003F94))
        self.reset_all_masks(process)

        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeAll
        )
        self.assertEqual(0x000002950001F694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFE950000F694, process.FixAddress(0xFFA65E950000F694))
        self.reset_all_masks(process)

        # Set a eAddressMaskTypeCode which has the low 3 bits marked as non-address
        # bits, confirm that they're cleared by FixAddress.
        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeAll
        )
        mask = process.GetAddressMask(lldb.eAddressMaskTypeAny)
        process.SetAddressMask(lldb.eAddressMaskTypeCode, mask | 0x3)
        self.assertEqual(0x000002950001F697, process.FixAddress(0x00265E950001F697))
        self.assertEqual(0xFFFFFE950000F697, process.FixAddress(0xFFA65E950000F697))
        self.assertEqual(
            0x000002950001F697,
            process.FixAddress(0x00265E950001F697, lldb.eAddressMaskTypeData),
        )
        self.assertEqual(
            0x000002950001F694,
            process.FixAddress(0x00265E950001F697, lldb.eAddressMaskTypeCode),
        )
        self.reset_all_masks(process)

        # The user can override whatever settings the Process thinks should be used.
        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeLow
        )
        self.runCmd("settings set target.process.virtual-addressable-bits 15")
        self.assertEqual(0x0000000000007694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFFFFFFFFF694, process.FixAddress(0xFFA65E950000F694))
        self.runCmd("settings set target.process.virtual-addressable-bits 0")
        self.assertEqual(0x000002950001F694, process.FixAddress(0x00265E950001F694))
        self.reset_all_masks(process)

    # AArch64 can have different address masks for high and low memory, when different
    # page tables are set up.
    @skipIf(archs=no_match(["arm64", "arm64e", "aarch64"]))
    @skipIf(archs=["arm$"])  # 32-bit arm ABI hardcodes Code mask, is 32-bit
    def test_address_masks_target_supports_highmem_tests(self):
        self.build()
        (target, process, t, bp) = lldbutil.run_to_source_breakpoint(
            self, "break here", lldb.SBFileSpec("main.c")
        )

        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeLow
        )
        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 15, lldb.eAddressMaskRangeHigh
        )
        self.assertEqual(0x000002950001F694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFFFFFFFFF694, process.FixAddress(0xFFA65E950000F694))
        self.reset_all_masks(process)

        # The user can override whatever settings the Process thinks should be used.
        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeAll
        )
        self.runCmd("settings set target.process.virtual-addressable-bits 15")
        self.runCmd("settings set target.process.highmem-virtual-addressable-bits 15")
        self.assertEqual(0x0000000000007694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFFFFFFFFF694, process.FixAddress(0xFFA65E950000F694))
        self.runCmd("settings set target.process.virtual-addressable-bits 0")
        self.runCmd("settings set target.process.highmem-virtual-addressable-bits 0")
        self.assertEqual(0x000002950001F694, process.FixAddress(0x00265E950001F694))
        self.reset_all_masks(process)

    # On most targets where we have a single mask for all address range, confirm
    # that the high memory masks are ignored.
    @skipIf(archs=["arm64", "arm64e", "aarch64"])
    @skipIf(archs=["arm$"])  # 32-bit arm ABI hardcodes Code mask, is 32-bit
    def test_address_masks_target_no_highmem(self):
        self.build()
        (target, process, t, bp) = lldbutil.run_to_source_breakpoint(
            self, "break here", lldb.SBFileSpec("main.c")
        )

        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 42, lldb.eAddressMaskRangeLow
        )
        process.SetAddressableBits(
            lldb.eAddressMaskTypeAll, 15, lldb.eAddressMaskRangeHigh
        )
        self.assertEqual(0x000002950001F694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFE950000F694, process.FixAddress(0xFFA65E950000F694))
        self.runCmd("settings set target.process.virtual-addressable-bits 15")
        self.runCmd("settings set target.process.highmem-virtual-addressable-bits 42")
        self.assertEqual(0x0000000000007694, process.FixAddress(0x00265E950001F694))
        self.assertEqual(0xFFFFFFFFFFFFF694, process.FixAddress(0xFFA65E950000F694))



# Formatters for classes that derive from Message.
#
# Usage:
#   command script import ./example.py
#   type summary add --expand --recognizer-function --python-function example.message_summary example.is_message_type
#   type synth add --recognizer-function --python-class example.MessageChildProvider example.is_message_type

import sys


def is_message_type(t, internal_dict):
    for base in t.get_bases_array():
        if base.GetName() == "Message":
            return True
    return False


def message_summary(value, internal_dict):
    # Could have used a summary string as well. All the work is done by the child
    # provider.
    return "Message"


class MessageChildProvider:
    def __init__(self, value, internal_dict):
        self.value = value
        self.synthetic_children = self._analyze_children(value)

    def has_children(self):
        return self.num_children() > 0

    def num_children(self):
        return len(self.synthetic_children)

    def get_child_index(self, name):
        for index, child in enumerate(self.synthetic_children):
            if child.GetName() == name:
                return index
        return None

    def get_child_at_index(self, index):
        return self.synthetic_children[index]

    def _rename_sbvalue(self, value):
        # We want to display the field with its original name without a trailing
        # underscore. So we create a new SBValue with the same type and address but
        # a different name.
        name = value.GetName()
        assert name.endswith("_")
        new_name = name[:-1]
        return value.CreateValueFromAddress(
            new_name, value.GetLoadAddress(), value.GetType()
        )

    def _analyze_children(self, value):
        result = []
        for i in range(value.GetNumChildren()):
            child = value.GetChildAtIndex(i)
            child_name = child.GetName()
            if child_name.startswith("_"):
                continue  # Internal field, skip
            # Normal field. Check presence bit.
            presence_bit = value.GetChildMemberWithName("_has_" + child_name)
            if presence_bit.GetValueAsUnsigned() != 0:
                result.append(self._rename_sbvalue(child))
        return result



"""
Test calling user defined functions using expression evaluation.
This test checks that typesystem lookup works correctly for typedefs of
untagged structures.

Ticket: https://llvm.org/bugs/show_bug.cgi?id=26790
"""

import lldb

from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestExprLookupAnonStructTypedef(TestBase):
    def test(self):
        """Test typedeffed untagged struct arguments for function call expressions"""
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "// break here", lldb.SBFileSpec("main.cpp")
        )
        self.expect_expr("multiply(&s)", result_type="double", result_value="1")



"""
Test that we handle breakpoints on consecutive instructions correctly.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class ConsecutiveBreakpointsTestCase(TestBase):
    def prepare_test(self):
        self.build()

        (
            self.target,
            self.process,
            self.thread,
            bkpt,
        ) = lldbutil.run_to_source_breakpoint(
            self, "Set breakpoint here", lldb.SBFileSpec("main.cpp")
        )

        # Set breakpoint to the next instruction
        frame = self.thread.GetFrameAtIndex(0)

        address = frame.GetPCAddress()
        instructions = self.target.ReadInstructions(address, 2)
        self.assertEqual(len(instructions), 2)
        self.bkpt_address = instructions[1].GetAddress()
        self.breakpoint2 = self.target.BreakpointCreateByAddress(
            self.bkpt_address.GetLoadAddress(self.target)
        )
        self.assertTrue(
            self.breakpoint2 and self.breakpoint2.GetNumLocations() == 1,
            VALID_BREAKPOINT,
        )

    def finish_test(self):
        # Run the process until termination
        self.process.Continue()
        self.assertState(self.process.GetState(), lldb.eStateExited)

    @no_debug_info_test
    def test_continue(self):
        """Test that continue stops at the second breakpoint."""
        self.prepare_test()

        self.process.Continue()
        self.assertState(self.process.GetState(), lldb.eStateStopped)
        # We should be stopped at the second breakpoint
        self.thread = lldbutil.get_one_thread_stopped_at_breakpoint(
            self.process, self.breakpoint2
        )
        self.assertIsNotNone(
            self.thread, "Expected one thread to be stopped at breakpoint 2"
        )

        self.finish_test()

    @no_debug_info_test
    def test_single_step(self):
        """Test that single step stops at the second breakpoint."""
        self.prepare_test()

        # Instruction step to the next instruction
        # We haven't executed breakpoint2 yet, we're sitting at it now.
        step_over = False
        self.thread.StepInstruction(step_over)

        step_over = False
        self.thread.StepInstruction(step_over)

        # We've now hit the breakpoint and this StepInstruction has
        # been interrupted, it is still sitting on the thread plan stack.

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertEqual(
            self.thread.GetFrameAtIndex(0).GetPCAddress().GetLoadAddress(self.target),
            self.bkpt_address.GetLoadAddress(self.target),
        )

        # One more instruction to complete the Step that was interrupted
        # earlier.
        self.thread.StepInstruction(step_over)
        strm = lldb.SBStream()
        self.thread.GetDescription(strm)
        self.assertIn("instruction step into", strm.GetData())
        self.assertIsNotNone(self.thread, "Expected to see that step-in had completed")

        self.finish_test()

    @no_debug_info_test
    @skipIf(
        oslist=["windows"],
        archs=["x86_64"],
        bugnumber="github.com/llvm/llvm-project/issues/138083",
    )
    def test_single_step_thread_specific(self):
        """Test that single step stops, even though the second breakpoint is not valid."""
        self.prepare_test()

        # Choose a thread other than the current one. A non-existing thread is
        # fine.
        thread_index = self.process.GetNumThreads() + 1
        self.assertFalse(self.process.GetThreadAtIndex(thread_index).IsValid())
        self.breakpoint2.SetThreadIndex(thread_index)

        step_over = False
        self.thread.StepInstruction(step_over)

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertEqual(
            self.thread.GetFrameAtIndex(0).GetPCAddress().GetLoadAddress(self.target),
            self.bkpt_address.GetLoadAddress(self.target),
        )
        self.assertEqual(
            self.thread.GetStopReason(),
            lldb.eStopReasonPlanComplete,
            "Stop reason should be 'plan complete'",
        )

        # Hit our second breakpoint
        self.process.Continue()

        self.finish_test()



"""
Test getting return-values correctly when stepping out
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class ReturnValueTestCase(TestBase):
    def affected_by_pr33042(self):
        return (
            "clang" in self.getCompiler()
            and self.isAArch64()
            and self.getPlatform() == "linux"
        )

    def affected_by_pr44132(self):
        return self.getArchitecture() in ["aarch64", "arm"] and self.getPlatform() in [
            "freebsd",
            "linux",
        ]

    # ABIMacOSX_arm64 and the SysV_arm64 don't restore the storage value for memory returns on function
    # exit, so lldb shouldn't attempt to fetch memory for those return types, as there is
    # no easy way to guarantee that they will be correct.  This is a list of the memory
    # return functions defined in the test file:
    arm_no_return_values = [
        "return_five_int",
        "return_one_int_one_double_one_int",
        "return_one_short_one_double_one_short",
        "return_vector_size_float32_32",
        "return_ext_vector_size_float32_8",
    ]

    def should_report_return_value(self, func_name):
        abi = self.target.GetABIName()
        if not abi in ["SysV-arm64", "ABIMacOSX_arm64", "macosx-arm"]:
            return True
        return not func_name in self.arm_no_return_values

    @expectedFailureAll(
        oslist=["freebsd"], archs=["i386"], bugnumber="llvm.org/pr48376"
    )
    @expectedFailureAll(
        oslist=["macosx"], archs=["i386"], bugnumber="<rdar://problem/28719652>"
    )
    @expectedFailureAll(
        oslist=["linux"],
        compiler="clang",
        compiler_version=["<=", "3.6"],
        archs=["i386"],
    )
    @expectedFailureAll(oslist=["windows"], bugnumber="llvm.org/pr24778")
    @add_test_categories(["pyapi"])
    def test_with_python(self):
        """Test getting return values from stepping out."""
        self.build()
        exe = self.getBuildArtifact("a.out")
        (
            self.target,
            self.process,
            thread,
            inner_sint_bkpt,
        ) = lldbutil.run_to_name_breakpoint(self, "inner_sint", exe_name=exe)

        error = lldb.SBError()

        # inner_sint returns the variable value, so capture that here:
        in_int = thread.GetFrameAtIndex(0).FindVariable("value").GetValueAsSigned(error)
        self.assertSuccess(error)

        thread.StepOut()

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertStopReason(thread.GetStopReason(), lldb.eStopReasonPlanComplete)

        frame = thread.GetFrameAtIndex(0)
        fun_name = frame.GetFunctionName()
        self.assertEqual(fun_name, "outer_sint(int)")

        return_value = thread.GetStopReturnValue()
        self.assertTrue(return_value.IsValid())

        ret_int = return_value.GetValueAsSigned(error)
        self.assertSuccess(error)
        self.assertEqual(in_int, ret_int)

        # Run again and we will stop in inner_sint the second time outer_sint is called.
        # Then test stepping out two frames at once:

        thread_list = lldbutil.continue_to_breakpoint(self.process, inner_sint_bkpt)
        self.assertEqual(len(thread_list), 1)
        thread = thread_list[0]

        # We are done with the inner_sint breakpoint:
        self.target.BreakpointDelete(inner_sint_bkpt.GetID())

        frame = thread.GetFrameAtIndex(1)
        fun_name = frame.GetFunctionName()
        self.assertEqual(fun_name, "outer_sint(int)")
        in_int = frame.FindVariable("value").GetValueAsSigned(error)
        self.assertSuccess(error)

        thread.StepOutOfFrame(frame)

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertStopReason(thread.GetStopReason(), lldb.eStopReasonPlanComplete)
        frame = thread.GetFrameAtIndex(0)
        fun_name = frame.GetFunctionName()
        self.assertEqual(fun_name, "main")

        ret_value = thread.GetStopReturnValue()
        self.assertTrue(return_value.IsValid())
        ret_int = ret_value.GetValueAsSigned(error)
        self.assertSuccess(error)
        self.assertEqual(2 * in_int, ret_int)

        # Now try some simple returns that have different types:
        inner_float_bkpt = self.target.BreakpointCreateByName("inner_float(float)", exe)
        self.assertTrue(inner_float_bkpt, VALID_BREAKPOINT)
        self.process.Continue()
        thread_list = lldbutil.get_threads_stopped_at_breakpoint(
            self.process, inner_float_bkpt
        )
        self.assertEqual(len(thread_list), 1)
        thread = thread_list[0]

        self.target.BreakpointDelete(inner_float_bkpt.GetID())

        frame = thread.GetFrameAtIndex(0)
        in_value = frame.FindVariable("value")
        in_float = float(in_value.GetValue())
        thread.StepOut()

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertStopReason(thread.GetStopReason(), lldb.eStopReasonPlanComplete)

        frame = thread.GetFrameAtIndex(0)
        fun_name = frame.GetFunctionName()
        self.assertEqual(fun_name, "outer_float(float)")

        # return_value = thread.GetStopReturnValue()
        # self.assertTrue(return_value.IsValid())
        # return_float = float(return_value.GetValue())

        # self.assertEqual(in_float, return_float)

        if not self.affected_by_pr44132():
            self.return_and_test_struct_value("return_one_int")
            self.return_and_test_struct_value("return_two_int")
            self.return_and_test_struct_value("return_three_int")
            self.return_and_test_struct_value("return_four_int")
            if not self.affected_by_pr33042():
                self.return_and_test_struct_value("return_five_int")
            self.return_and_test_struct_value("return_two_double")
            self.return_and_test_struct_value("return_one_double_two_float")
            self.return_and_test_struct_value("return_one_int_one_float_one_int")

            self.return_and_test_struct_value("return_one_pointer")
            self.return_and_test_struct_value("return_two_pointer")
            self.return_and_test_struct_value("return_one_float_one_pointer")
            self.return_and_test_struct_value("return_one_int_one_pointer")
            self.return_and_test_struct_value("return_three_short_one_float")

            self.return_and_test_struct_value("return_one_int_one_double")
            self.return_and_test_struct_value("return_one_int_one_double_one_int")
            self.return_and_test_struct_value("return_one_short_one_double_one_short")
            self.return_and_test_struct_value("return_one_float_one_int_one_float")
            self.return_and_test_struct_value("return_two_float")
            # I am leaving out the packed test until we have a way to tell CLANG
            # about alignment when reading DWARF for packed types.
            # self.return_and_test_struct_value ("return_one_int_one_double_packed")
            self.return_and_test_struct_value("return_one_int_one_long")

    @expectedFailureAll(
        oslist=["freebsd"], archs=["i386"], bugnumber="llvm.org/pr48376"
    )
    @expectedFailureAll(
        oslist=["macosx"], archs=["i386"], bugnumber="<rdar://problem/28719652>"
    )
    @expectedFailureAll(
        oslist=["linux"],
        compiler="clang",
        compiler_version=["<=", "3.6"],
        archs=["i386"],
    )
    @expectedFailureAll(compiler=["gcc"], archs=["x86_64", "i386"])
    @expectedFailureAll(
        oslist=["windows"], archs=["i[3-6]86", "x86_64"], bugnumber="llvm.org/pr24778"
    )
    def test_vector_values(self):
        self.build()
        exe = self.getBuildArtifact("a.out")
        error = lldb.SBError()

        self.target = self.dbg.CreateTarget(exe)
        self.assertTrue(self.target, VALID_TARGET)

        main_bktp = self.target.BreakpointCreateByName("main", exe)
        self.assertTrue(main_bktp, VALID_BREAKPOINT)

        self.process = self.target.LaunchSimple(
            None, None, self.get_process_working_directory()
        )
        self.assertEqual(
            len(lldbutil.get_threads_stopped_at_breakpoint(self.process, main_bktp)), 1
        )
        self.return_and_test_struct_value("return_vector_size_float32_8")
        self.return_and_test_struct_value("return_vector_size_float32_16")
        if not self.affected_by_pr44132():
            self.return_and_test_struct_value("return_vector_size_float32_32")
        self.return_and_test_struct_value("return_ext_vector_size_float32_2")
        self.return_and_test_struct_value("return_ext_vector_size_float32_4")
        if not self.affected_by_pr44132():
            self.return_and_test_struct_value("return_ext_vector_size_float32_8")

    # limit the nested struct and class tests to only x86_64
    @skipIf(archs=no_match(["x86_64"]))
    @expectedFailureAll(oslist=["windows"], bugnumber="llvm.org/pr24778")
    def test_for_cpp_support(self):
        self.build()
        exe = self.getBuildArtifact("a.out")
        (
            self.target,
            self.process,
            thread,
            inner_sint_bkpt,
        ) = lldbutil.run_to_name_breakpoint(self, "inner_sint", exe_name=exe)

        error = lldb.SBError()

        self.target = self.dbg.CreateTarget(exe)
        self.assertTrue(self.target, VALID_TARGET)

        main_bktp = self.target.BreakpointCreateByName("main", exe)
        self.assertTrue(main_bktp, VALID_BREAKPOINT)

        self.process = self.target.LaunchSimple(
            None, None, self.get_process_working_directory()
        )
        self.assertEqual(
            len(lldbutil.get_threads_stopped_at_breakpoint(self.process, main_bktp)), 1
        )
        # nested struct tests
        self.return_and_test_struct_value("return_nested_one_float_three_base")
        self.return_and_test_struct_value("return_double_nested_one_float_one_nested")
        self.return_and_test_struct_value("return_nested_float_struct")
        # class test
        self.return_and_test_struct_value("return_base_class_one_char")
        self.return_and_test_struct_value("return_nested_class_float_and_base")
        self.return_and_test_struct_value("return_double_nested_class_float_and_nested")
        self.return_and_test_struct_value("return_base_class")
        self.return_and_test_struct_value("return_derived_class")

    @skipIf(compiler="clang", compiler_version=["<", "7.0"])
    def return_and_test_struct_value(self, func_name):
        """Pass in the name of the function to return from - takes in value, returns value."""

        # Set the breakpoint, run to it, finish out.
        bkpt = self.target.BreakpointCreateByName(func_name)
        self.assertGreater(
            bkpt.GetNumResolvedLocations(),
            0,
            "Got wrong number of locations for {0}".format(func_name),
        )

        self.process.Continue()

        thread_list = lldbutil.get_threads_stopped_at_breakpoint(self.process, bkpt)

        self.assertEqual(len(thread_list), 1)
        thread = thread_list[0]

        self.target.BreakpointDelete(bkpt.GetID())

        in_value = thread.GetFrameAtIndex(0).FindVariable("value")

        self.assertTrue(in_value.IsValid())
        num_in_children = in_value.GetNumChildren()

        # This is a little hokey, but if we don't get all the children now, then
        # once we've stepped we won't be able to get them?

        for idx in range(0, num_in_children):
            in_child = in_value.GetChildAtIndex(idx)
            in_child_str = in_child.GetValue()

        thread.StepOut()

        self.assertState(self.process.GetState(), lldb.eStateStopped)
        self.assertStopReason(thread.GetStopReason(), lldb.eStopReasonPlanComplete)

        # Assuming all these functions step out to main.  Could figure out the caller dynamically
        # if that would add something to the test.
        frame = thread.GetFrameAtIndex(0)
        fun_name = frame.GetFunctionName()
        self.assertEqual(fun_name, "main")

        frame = thread.GetFrameAtIndex(0)
        ret_value = thread.GetStopReturnValue()
        if not self.should_report_return_value(func_name):
            self.assertFalse(ret_value.IsValid(), "Shouldn't have gotten a value")
            return

        self.assertTrue(ret_value.IsValid())

        num_ret_children = ret_value.GetNumChildren()
        self.assertEqual(num_in_children, num_ret_children)
        for idx in range(0, num_ret_children):
            in_child = in_value.GetChildAtIndex(idx)
            ret_child = ret_value.GetChildAtIndex(idx)
            in_child_str = in_child.GetValue()
            ret_child_str = ret_child.GetValue()

            self.assertEqual(in_child_str, ret_child_str)



# RUN: %PYTHON %s | FileCheck %s

from mlir.ir import *
from mlir.passmanager import *

from mlir.dialects import sparse_tensor as st


def run(f):
    print("\nTEST:", f.__name__)
    f()
    return f


# CHECK-LABEL: TEST: testSparseTensorPass
@run
def testSparseTensorPass():
    with Context() as context:
        PassManager.parse("any(sparsification)")
        PassManager.parse("any(sparse-tensor-conversion)")
    # CHECK: SUCCESS
    print("SUCCESS")



"""
Test the "process continue --reverse" and "--forward" options.
"""


import lldb
from lldbsuite.test.lldbtest import *
from lldbsuite.test.decorators import *
from lldbsuite.test.gdbclientutils import *
from lldbsuite.test.lldbreverse import ReverseTestBase
from lldbsuite.test import lldbutil


class TestReverseContinue(ReverseTestBase):
    @skipIfRemote
    def test_reverse_continue(self):
        target, _, _ = self.setup_recording()

        # Set breakpoint and reverse-continue
        trigger_bkpt = target.BreakpointCreateByName("trigger_breakpoint", None)
        self.assertTrue(trigger_bkpt.GetNumLocations() > 0)
        self.expect(
            "process continue --reverse",
            substrs=["stop reason = breakpoint {0}.1".format(trigger_bkpt.GetID())],
        )
        # `process continue` should preserve current base direction.
        self.expect(
            "process continue",
            STOPPED_DUE_TO_HISTORY_BOUNDARY,
            substrs=["stopped", "stop reason = history boundary"],
        )
        self.expect(
            "process continue --forward",
            substrs=["stop reason = breakpoint {0}.1".format(trigger_bkpt.GetID())],
        )

    def setup_recording(self):
        """
        Record execution of code between "start_recording" and "stop_recording" breakpoints.

        Returns with the target stopped at "stop_recording", with recording disabled,
        ready to reverse-execute.
        """
        self.build()
        target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))
        process = self.connect(target)

        # Record execution from the start of the function "start_recording"
        # to the start of the function "stop_recording". We want to keep the
        # interval that we record as small as possible to minimize the run-time
        # of our single-stepping recorder.
        start_recording_bkpt = target.BreakpointCreateByName("start_recording", None)
        self.assertTrue(start_recording_bkpt.GetNumLocations() > 0)
        initial_threads = lldbutil.continue_to_breakpoint(process, start_recording_bkpt)
        self.assertEqual(len(initial_threads), 1)
        target.BreakpointDelete(start_recording_bkpt.GetID())
        self.start_recording()
        stop_recording_bkpt = target.BreakpointCreateByName("stop_recording", None)
        self.assertTrue(stop_recording_bkpt.GetNumLocations() > 0)
        lldbutil.continue_to_breakpoint(process, stop_recording_bkpt)
        target.BreakpointDelete(stop_recording_bkpt.GetID())
        self.stop_recording()

        self.dbg.SetAsync(False)

        return target, process, initial_threads



"""
Verify the default cache line size for android targets
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class DefaultCacheLineSizeTestCase(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    @skipUnlessTargetAndroid
    def test_cache_line_size(self):
        self.build()
        target = self.createTestTarget()
        self.assertTrue(target and target.IsValid(), "Target is valid")

        breakpoint = target.BreakpointCreateByName("main")
        self.assertTrue(breakpoint and breakpoint.IsValid(), "Breakpoint is valid")

        # Run the program.
        process = target.LaunchSimple(None, None, self.get_process_working_directory())
        self.assertTrue(process and process.IsValid(), PROCESS_IS_VALID)
        self.assertEqual(process.GetState(), lldb.eStateStopped, PROCESS_STOPPED)

        # check the setting value
        self.expect(
            "settings show target.process.memory-cache-line-size", patterns=[" = 2048"]
        )

        # Run to completion.
        process.Continue()
        self.assertState(process.GetState(), lldb.eStateExited, PROCESS_EXITED)



"""
Test lldb breakpoint setting by source regular expression.
This test just tests the source file & function restrictions.
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestSourceRegexBreakpoints(TestBase):
    def test_location(self):
        self.build()
        self.source_regex_locations()

    def test_restrictions(self):
        self.build()
        self.source_regex_restrictions()

    def source_regex_locations(self):
        """Test that restricting source expressions to files & to functions."""
        # Create a target by the debugger.
        exe = self.getBuildArtifact("a.out")
        target = self.dbg.CreateTarget(exe)
        self.assertTrue(target, VALID_TARGET)

        # First look just in main:
        target_files = lldb.SBFileSpecList()
        target_files.Append(lldb.SBFileSpec("a.c"))

        func_names = lldb.SBStringList()
        func_names.AppendString("a_func")

        source_regex = "Set . breakpoint here"
        main_break = target.BreakpointCreateBySourceRegex(
            source_regex, lldb.SBFileSpecList(), target_files, func_names
        )
        num_locations = main_break.GetNumLocations()
        self.assertEqual(
            num_locations,
            1,
            "a.c in a_func should give one breakpoint, got %d." % (num_locations),
        )

        loc = main_break.GetLocationAtIndex(0)
        self.assertTrue(loc.IsValid(), "Got a valid location.")
        address = loc.GetAddress()
        self.assertTrue(address.IsValid(), "Got a valid address from the location.")

        a_func_line = line_number("a.c", "Set A breakpoint here")
        line_entry = address.GetLineEntry()
        self.assertTrue(line_entry.IsValid(), "Got a valid line entry.")
        self.assertEqual(
            line_entry.line,
            a_func_line,
            "Our line number matches the one lldbtest found.",
        )

    def source_regex_restrictions(self):
        """Test that restricting source expressions to files & to functions."""
        # Create a target by the debugger.
        exe = self.getBuildArtifact("a.out")
        target = self.dbg.CreateTarget(exe)
        self.assertTrue(target, VALID_TARGET)

        # First look just in main:
        target_files = lldb.SBFileSpecList()
        target_files.Append(lldb.SBFileSpec("main.c"))
        source_regex = "Set . breakpoint here"
        main_break = target.BreakpointCreateBySourceRegex(
            source_regex, lldb.SBFileSpecList(), target_files, lldb.SBStringList()
        )

        num_locations = main_break.GetNumLocations()
        self.assertEqual(
            num_locations, 2, "main.c should have 2 matches, got %d." % (num_locations)
        )

        # Now look in both files:
        target_files.Append(lldb.SBFileSpec("a.c"))

        main_break = target.BreakpointCreateBySourceRegex(
            source_regex, lldb.SBFileSpecList(), target_files, lldb.SBStringList()
        )

        num_locations = main_break.GetNumLocations()
        self.assertEqual(
            num_locations,
            4,
            "main.c and a.c should have 4 matches, got %d." % (num_locations),
        )

        # Now restrict it to functions:
        func_names = lldb.SBStringList()
        func_names.AppendString("main_func")
        main_break = target.BreakpointCreateBySourceRegex(
            source_regex, lldb.SBFileSpecList(), target_files, func_names
        )

        num_locations = main_break.GetNumLocations()
        self.assertEqual(
            num_locations,
            2,
            "main_func in main.c and a.c should have 2 matches, got %d."
            % (num_locations),
        )



"""
Fuzz tests an object after the default construction to make sure it does not crash lldb.
"""

import sys
import lldb


def fuzz_obj(obj):
    obj.GetFileAddress()
    obj.GetLoadAddress(lldb.SBTarget())
    obj.SetLoadAddress(0xFFFF, lldb.SBTarget())
    obj.OffsetAddress(sys.maxsize)
    obj.GetDescription(lldb.SBStream())
    obj.GetSection()
    obj.GetSymbolContext(lldb.eSymbolContextEverything)
    obj.GetModule()
    obj.GetCompileUnit()
    obj.GetFunction()
    obj.GetBlock()
    obj.GetSymbol()
    obj.GetLineEntry()
    obj.Clear()



"""
Test the IR interpreter
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class IRInterpreterTestCase(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def time_expression(self, expr, options):
        start = time.time()
        res = self.target.EvaluateExpression(expr, options)
        return res, time.time() - start

    def test_interpreter_timeout(self):
        """Test the timeout logic in the IRInterpreter."""
        self.build()
        self.target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))
        self.assertTrue(self.target, VALID_TARGET)

        # A non-trivial infinite loop.
        inf_loop = "for (unsigned i = 0; i < 100; ++i) --i; 1"
        timeout_error = "Reached timeout while interpreting expression"
        options = lldb.SBExpressionOptions()

        # This is an IRInterpreter specific test, so disable the JIT.
        options.SetAllowJIT(False)

        # We use a 500ms timeout.
        options.SetTimeoutInMicroSeconds(500000)
        res, duration_sec = self.time_expression(inf_loop, options)
        self.assertIn(timeout_error, str(res.GetError()))

        # Depending on the machine load the expression might take quite some
        # time, so give the time a generous upper bound.
        self.assertLess(duration_sec, 15)

        # Try a simple one second timeout.
        options.SetTimeoutInMicroSeconds(1000000)
        res, duration_sec = self.time_expression(inf_loop, options)
        self.assertIn(timeout_error, str(res.GetError()))
        # Anything within 5% of 1s is fine, to account for machine differences.
        self.assertGreaterEqual(duration_sec, 0.95)
        self.assertLess(duration_sec, 30)

    def test_interpreter_interrupt(self):
        """Test interrupting the IRInterpreter."""
        self.build()
        self.target = self.dbg.CreateTarget(self.getBuildArtifact("a.out"))
        self.assertTrue(self.target, VALID_TARGET)

        # A non-trivial infinite loop.
        inf_loop = "for (unsigned i = 0; i < 100; ++i) --i; 1"

        options = lldb.SBExpressionOptions()

        # This is an IRInterpreter specific test, so disable the JIT.
        options.SetAllowJIT(False)

        # Make sure we have a pretty long (10s) timeout so we have a chance to
        # interrupt the interpreted expression.
        options.SetTimeoutInMicroSeconds(10000000)

        self.dbg.RequestInterrupt()

        self.dbg.SetAsync(True)
        res = self.target.EvaluateExpression(inf_loop, options)
        self.dbg.SetAsync(False)

        # Be sure to turn this off again:
        def cleanup():
            if self.dbg.InterruptRequested():
                self.dbg.CancelInterruptRequest()

        self.addTearDownHook(cleanup)

        interrupt_error = "Interrupted while interpreting expression"
        self.assertIn(interrupt_error, str(res.GetError()))

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Find the line number to break for main.c.
        self.line = line_number("main.c", "// Set breakpoint here")

        # Disable confirmation prompt to avoid infinite wait
        self.runCmd("settings set auto-confirm true")
        self.addTearDownHook(lambda: self.runCmd("settings clear auto-confirm"))

    def build_and_run(self):
        """Test the IR interpreter"""
        self.build()

        self.runCmd("file " + self.getBuildArtifact("a.out"), CURRENT_EXECUTABLE_SET)

        lldbutil.run_break_set_by_file_and_line(
            self, "main.c", self.line, num_expected_locations=1, loc_exact=False
        )

        self.runCmd("run", RUN_SUCCEEDED)

    @expectedFailureAll(
        oslist=["windows"],
        bugnumber="llvm.org/pr21765",
    )
    @add_test_categories(["pyapi"])
    def test_ir_interpreter(self):
        self.build_and_run()

        options = lldb.SBExpressionOptions()
        options.SetLanguage(lldb.eLanguageTypeC_plus_plus)

        set_up_expressions = [
            "int $i = 9",
            "int $j = 3",
            "int $k = 5",
            "unsigned long long $ull = -1",
            "unsigned $u = -1",
        ]

        expressions = [
            "$i + $j",
            "$i - $j",
            "$i * $j",
            "$i / $j",
            "$i % $k",
            "$i << $j",
            "$i & $j",
            "$i | $j",
            "$i ^ $j",
            "($ull & -1) == $u",
        ]

        for expression in set_up_expressions:
            self.frame().EvaluateExpression(expression, options)

        func_call = "(int)getpid()"
        if lldbplatformutil.getPlatform() == "windows":
            func_call = "(int)GetCurrentProcessId()"

        for expression in expressions:
            interp_expression = expression
            jit_expression = func_call + "; " + expression

        for expression in expressions:
            interp_expression = expression
            jit_expression = "(int)getpid(); " + expression

            interp_result = (
                self.frame()
                .EvaluateExpression(interp_expression, options)
                .GetValueAsSigned()
            )
            jit_result = (
                self.frame()
                .EvaluateExpression(jit_expression, options)
                .GetValueAsSigned()
            )

            self.assertEqual(
                interp_result, jit_result, "While evaluating " + expression
            )

    def test_type_conversions(self):
        target = self.dbg.GetDummyTarget()
        short_val = target.EvaluateExpression("(short)-1")
        self.assertEqual(short_val.GetValueAsSigned(), -1)
        long_val = target.EvaluateExpression("(long) " + short_val.GetName())
        self.assertEqual(long_val.GetValueAsSigned(), -1)

    def test_fpconv(self):
        self.build_and_run()

        interp_options = lldb.SBExpressionOptions()
        interp_options.SetLanguage(lldb.eLanguageTypeC_plus_plus)
        interp_options.SetAllowJIT(False)

        jit_options = lldb.SBExpressionOptions()
        jit_options.SetLanguage(lldb.eLanguageTypeC_plus_plus)
        jit_options.SetAllowJIT(True)

        set_up_expressions = [
            "int32_t $i = 3",
            "int32_t $n = -3",
            "uint32_t $u = 5",
            "int64_t $l = -7",
            "float $f = 9.0625",
            "double $d = 13.75",
            "float $nf = -11.25",
        ]

        expressions = [
            "$i + $f",  # sitofp i32 to float
            "$d - $n",  # sitofp i32 to double
            "$u + $f",  # uitofp i32 to float
            "$u + $d",  # uitofp i32 to double
            "(int32_t)$d",  # fptosi double to i32
            "(int32_t)$f",  # fptosi float to i32
            "(int64_t)$d",  # fptosi double to i64
            "(int16_t)$f",  # fptosi float to i16
            "(int64_t)$nf",  # fptosi float to i64
            "(uint16_t)$f",  # fptoui float to i16
            "(uint32_t)$d",  # fptoui double to i32
            "(uint64_t)$d",  # fptoui double to i64
            "(float)$d",  # fptrunc double to float
            "(double)$f",  # fpext float to double
            "(double)$nf",  # fpext float to double
        ]

        for expression in set_up_expressions:
            self.frame().EvaluateExpression(expression, interp_options)

        func_call = "(int)getpid()"
        if lldbplatformutil.getPlatform() == "windows":
            func_call = "(int)GetCurrentProcessId()"

        for expression in expressions:
            interp_expression = expression
            # Calling a function forces the expression to be executed with JIT.
            jit_expression = func_call + "; " + expression

            interp_result = self.frame().EvaluateExpression(
                interp_expression, interp_options
            )
            jit_result = self.frame().EvaluateExpression(jit_expression, jit_options)

            self.assertEqual(
                interp_result.GetValue(),
                jit_result.GetValue(),
                "Values match for " + expression,
            )
            self.assertEqual(
                interp_result.GetTypeName(),
                jit_result.GetTypeName(),
                "Types match for " + expression,
            )

    def test_fpconv_ub(self):
        target = self.dbg.GetDummyTarget()

        set_up_expressions = [
            "float $f = 3e9",
            "double $d = 1e20",
            "float $nf = -1.5",
        ]

        expressions = [
            (
                "(int32_t)$f",
                "Conversion error: (float) 3.0E+9 cannot be converted to i32",
            ),
            (
                "(uint32_t)$nf",
                "Conversion error: (float) -1.5 cannot be converted to unsigned i32",
            ),
            (
                "(int64_t)$d",
                "Conversion error: (float) 1.0E+20 cannot be converted to i64",
            ),
            (
                "(uint64_t)$d",
                "Conversion error: (float) 1.0E+20 cannot be converted to unsigned i64",
            ),
        ]

        for expression in set_up_expressions:
            target.EvaluateExpression(expression)

        # The IR Interpreter returns an error if a value cannot be converted.
        for expression in expressions:
            result = target.EvaluateExpression(expression[0])
            self.assertIn(expression[1], str(result.GetError()))

        # The conversion should succeed if the destination type can represent the result.
        self.expect_expr(
            "(uint32_t)$f", result_type="uint32_t", result_value="3000000000"
        )
        self.expect_expr("(int32_t)$nf", result_type="int32_t", result_value="-1")



class WrapperSynthProvider:
    def __init__(self, valobj, internal_dict):
        self.valobj = valobj
        self.sum_value = None

    def update(self):
        self.sum_value = None
        ty = self.valobj.GetType()

        # Artificially bail out if LLDB passed us a reference or a pointer.
        if ty.IsPointerType() or ty.IsReferenceType():
            return False

        x = self.valobj.GetChildMemberWithName("x")
        y = self.valobj.GetChildMemberWithName("y")
        if x.IsValid() and y.IsValid():
            sum_val = x.GetValueAsUnsigned(0) + y.GetValueAsUnsigned(0)
            self.sum_value = self.valobj.CreateValueFromExpression("sum", str(sum_val))

        return False

    def num_children(self):
        return 1

    def get_child_at_index(self, index):
        if index == 0 and self.sum_value:
            return self.sum_value
        return None

    def get_child_index(self, name):
        if name == "sum":
            return 0
        return -1

    def has_children(self):
        return True



"""
Test the AArch64 SME Array Storage (ZA) register dynamic resize with
multiple threads.
"""

from enum import Enum
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class AArch64ZAThreadedTestCase(TestBase):
    def get_supported_vg(self):
        exe = self.getBuildArtifact("a.out")
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        main_thread_stop_line = line_number("main.c", "// Break in main thread")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", main_thread_stop_line)

        self.runCmd("settings set target.run-args 0")
        self.runCmd("run", RUN_SUCCEEDED)

        self.expect(
            "thread info 1",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stop reason = breakpoint"],
        )

        current_vg = self.match("register read vg", ["(0x[0-9]+)"])
        self.assertIsNotNone(current_vg)
        self.expect("register write vg {}".format(current_vg.group()))

        # Aka 128, 256 and 512 bit.
        supported_vg = []
        for vg in [2, 4, 8]:
            # This could mask other errors but writing vg is tested elsewhere
            # so we assume the hardware rejected the value.
            self.runCmd("register write vg {}".format(vg), check=False)
            if not self.res.GetError():
                supported_vg.append(vg)

        self.runCmd("breakpoint delete 1")
        self.runCmd("continue")

        return supported_vg

    def gen_za_value(self, svg, value_generator):
        svl = svg * 8

        rows = []
        for row in range(svl):
            byte = "0x{:02x}".format(value_generator(row))
            rows.append(" ".join([byte] * svl))

        return "{" + " ".join(rows) + "}"

    def check_za_register(self, svg, value_offset):
        self.expect(
            "register read za",
            substrs=[self.gen_za_value(svg, lambda r: r + value_offset)],
        )

    def check_disabled_za_register(self, svg):
        self.expect("register read za", substrs=[self.gen_za_value(svg, lambda r: 0)])

    def za_test_impl(self, enable_za):
        # Although the test program doesn't obviously do any operations that
        # would need smefa64, calls to libc functions like memset may do.
        if not self.isAArch64SMEFA64():
            self.skipTest("SME and the sm3fa64 extension must be present")

        self.build()
        supported_vg = self.get_supported_vg()

        self.runCmd("settings set target.run-args {}".format("1" if enable_za else "0"))

        if not (2 in supported_vg and 4 in supported_vg):
            self.skipTest("Not all required streaming vector lengths are supported.")

        main_thread_stop_line = line_number("main.c", "// Break in main thread")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", main_thread_stop_line)

        thX_break_line1 = line_number("main.c", "// Thread X breakpoint 1")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", thX_break_line1)

        thX_break_line2 = line_number("main.c", "// Thread X breakpoint 2")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", thX_break_line2)

        thY_break_line1 = line_number("main.c", "// Thread Y breakpoint 1")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", thY_break_line1)

        thY_break_line2 = line_number("main.c", "// Thread Y breakpoint 2")
        lldbutil.run_break_set_by_file_and_line(self, "main.c", thY_break_line2)

        self.runCmd("run", RUN_SUCCEEDED)

        self.expect(
            "thread info 1",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stop reason = breakpoint"],
        )

        if 8 in supported_vg:
            if enable_za:
                self.check_za_register(8, 1)
            else:
                self.check_disabled_za_register(8)
        else:
            if enable_za:
                self.check_za_register(4, 1)
            else:
                self.check_disabled_za_register(4)

        self.runCmd("process continue", RUN_SUCCEEDED)

        process = self.dbg.GetSelectedTarget().GetProcess()
        for idx in range(1, process.GetNumThreads()):
            thread = process.GetThreadAtIndex(idx)
            if thread.GetStopReason() != lldb.eStopReasonBreakpoint:
                self.runCmd("thread continue %d" % (idx + 1))
                self.assertEqual(thread.GetStopReason(), lldb.eStopReasonBreakpoint)

            stopped_at_line_number = thread.GetFrameAtIndex(0).GetLineEntry().GetLine()

            if stopped_at_line_number == thX_break_line1:
                self.runCmd("thread select %d" % (idx + 1))
                self.check_za_register(4, 2)
                self.runCmd("register write vg 2")
                self.check_disabled_za_register(2)

            elif stopped_at_line_number == thY_break_line1:
                self.runCmd("thread select %d" % (idx + 1))
                self.check_za_register(2, 3)
                self.runCmd("register write vg 4")
                self.check_disabled_za_register(4)

        self.runCmd("thread continue 2")
        self.runCmd("thread continue 3")

        for idx in range(1, process.GetNumThreads()):
            thread = process.GetThreadAtIndex(idx)
            self.assertEqual(thread.GetStopReason(), lldb.eStopReasonBreakpoint)

            stopped_at_line_number = thread.GetFrameAtIndex(0).GetLineEntry().GetLine()

            if stopped_at_line_number == thX_break_line2:
                self.runCmd("thread select %d" % (idx + 1))
                self.check_za_register(2, 2)

            elif stopped_at_line_number == thY_break_line2:
                self.runCmd("thread select %d" % (idx + 1))
                self.check_za_register(4, 3)

    @no_debug_info_test
    @skipIf(archs=no_match(["aarch64"]))
    @skipIf(oslist=no_match(["linux"]))
    def test_za_register_dynamic_config_main_enabled(self):
        """Test multiple threads resizing ZA, with the main thread's ZA
        enabled."""
        self.za_test_impl(True)

    @no_debug_info_test
    @skipIf(archs=no_match(["aarch64"]))
    @skipIf(oslist=no_match(["linux"]))
    def test_za_register_dynamic_config_main_disabled(self):
        """Test multiple threads resizing ZA, with the main thread's ZA
        disabled."""
        self.za_test_impl(False)



import lldb


class FacadeExample:
    def __init__(self, bkpt, extra_args, dict):
        self.bkpt = bkpt
        self.extra_args = extra_args
        self.base_sym = None
        self.facade_locs = []
        self.facade_locs_desc = []
        self.cur_facade_loc = 1

        self.sym_name = extra_args.GetValueForKey("symbol").GetStringValue(100)
        self.num_locs = extra_args.GetValueForKey("num_locs").GetIntegerValue(5)
        self.loc_to_miss = extra_args.GetValueForKey("loc_to_miss").GetIntegerValue(
            10000
        )

    def __callback__(self, sym_ctx):
        self.base_sym = sym_ctx.module.FindSymbol(self.sym_name, lldb.eSymbolTypeCode)
        if self.base_sym.IsValid():
            self.bkpt.AddLocation(self.base_sym.GetStartAddress())
            # Locations are 1 based, so to keep things simple, I'm making
            # the array holding locations 1 based as well:
            self.facade_locs_desc.append(
                "This is the zero index, you shouldn't see this"
            )
            self.facade_locs.append(None)
            for i in range(1, self.num_locs + 1):
                self.facade_locs_desc.append(f"Location index: {i}")
                self.facade_locs.append(self.bkpt.AddFacadeLocation())

    def get_short_help(self):
        return f"I am a facade resolver - sym: {self.sym_name} - num_locs: {self.num_locs} - locs_to_miss: {self.loc_to_miss}"

    def was_hit(self, frame, bp_loc):
        tmp_loc = self.cur_facade_loc

        self.cur_facade_loc = self.cur_facade_loc + 1
        if self.cur_facade_loc == self.num_locs + 1:
            self.cur_facade_loc = 1

        if tmp_loc == self.loc_to_miss:
            return None

        # Make sure we can call a function here as well:
        options = lldb.SBExpressionOptions()
        options.SetStopOthers(True)
        options.SetTryAllThreads(False)

        result = frame.EvaluateExpression("change_him()", options)
        if not result.error.success:
            return lldb.LLDB_INVALID_BREAK_ID

        return self.facade_locs[tmp_loc]

    def get_location_description(self, bp_loc, desc_level):
        return self.facade_locs_desc[bp_loc.id]
